# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import Callable, Optional, Union

# Third Party
from grpc import StatusCode
from prometheus_client import Summary

# First Party
import alog

# Local
from caikit.config import get_config
from caikit.core import MODEL_MANAGER, ModuleBase
from caikit.runtime.model_management.batcher import Batcher
from caikit.runtime.model_management.loaded_model import LoadedModel
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from caikit.runtime.work_management.abortable_action import (
    AbortableAction,
    ActionAborter,
)

log = alog.use_channel("MODEL-LOADER")

CAIKIT_CORE_LOAD_DURATION_SUMMARY = Summary(
    "caikit_core_load_model_duration_seconds",
    "Summary of the duration (in seconds) of caikit.core.load(model)",
    ["model_type"],
)


class ModelLoader:
    """Model Loader class. The singleton class contains the core implementation details
    for loading models in from S3."""

    __instance = None

    def __init__(self):
        # Re-instantiating this is a programming error
        assert self.__class__.__instance is None, "This class is a singleton!"
        ModelLoader.__instance = self
        self._load_thread_pool = ThreadPoolExecutor(get_config().runtime.load_threads)
        # Instead of storing config-based batching information here, we call
        # get_config() when needed to support dynamic config changes for
        # batching

    def load_model(
        self,
        model_id: str,
        local_model_path: str,
        model_type: str,
        aborter: Optional[ActionAborter] = None,
        fail_callback: Optional[Callable] = None,
        retries: int = 0,
    ) -> LoadedModel:
        """Start loading a model from disk and associate the ID/size with it

        Args:
            model_id (str): Model ID string for the model to load.
            local_model_path (str): Local filesystem path to load the model from.
            model_type (str): Type of the model to load.
            aborter (Optional[ActionAborter]): An aborter to use that will allow
                the call's parent to abort the load
            fail_callback (Optional[Callable]): Optional no-arg callback to call
                on load failure
            retries (int): Number of times to retry loading
        Returns:
            model (LoadedModel) : The model that was loaded
        """
        # Set up the basics of the model's metadata
        model_builder = (
            LoadedModel.Builder()
            .id(model_id)
            .type(model_type)
            .path(local_model_path)
            .fail_callback(fail_callback)
            .retries(retries)
        )

        # Set up the async loading
        args = (local_model_path, model_id, model_type)
        log.debug2("Loading model %s async", model_id)
        if aborter is not None:
            log.debug3("Using abortable action to load %s", model_id)
            action = AbortableAction(aborter, self._load_module, *args)
            future_factory = partial(self._load_thread_pool.submit, action.do)
        else:
            future_factory = partial(
                self._load_thread_pool.submit, self._load_module, *args
            )
        model_builder.model_future_factory(future_factory)

        # Return the built model with the future handle
        return model_builder.build()

    def _load_module(
        self, model_path: str, model_id: str, model_type: str
    ) -> LoadedModel:
        try:
            log.info("<RUN89711114I>", "Loading model '%s'", model_id)

            # Load using the caikit.core
            with CAIKIT_CORE_LOAD_DURATION_SUMMARY.labels(model_type=model_type).time():
                model = MODEL_MANAGER.load(model_path)

            # If this model needs batching, configure a Batcher to wrap it
            model = self._wrap_in_batcher_if_configured(
                model,
                model_type,
                model_id,
            )
        except FileNotFoundError as fnfe:
            log_dict = {
                "log_code": "<RUN98613924E>",
                "message": "load failed to find model: %s with error: %s"
                % (model_path, repr(fnfe)),
                "model_id": model_id,
            }
            log.error(log_dict)
            raise CaikitRuntimeException(
                StatusCode.NOT_FOUND,
                f"Model {model_id} not found. Nested error: {fnfe}",
            ) from fnfe
        except Exception as ex:
            log_dict = {
                "log_code": "<RUN62912924E>",
                "message": "load failed when processing path: %s with error: %s"
                % (model_path, repr(ex)),
                "model_id": model_id,
            }
            log.error(log_dict, exc_info=True)
            raise CaikitRuntimeException(
                StatusCode.INTERNAL,
                f"Model {model_id} failed to load. Nested error: {ex}",
            ) from ex

        cache_info = MODEL_MANAGER.get_singleton_model_cache_info()
        log.info("<RUN89713784I>", "Singleton cache: '%s'", str(cache_info))

        return model

    @classmethod
    def get_instance(cls) -> "ModelLoader":
        """This method returns the instance of Model Manager"""
        if not cls.__instance:
            cls.__instance = ModelLoader()
        return cls.__instance

    def _wrap_in_batcher_if_configured(
        self,
        caikit_core_model: ModuleBase,
        model_type: str,
        model_id: str,
    ) -> Union[Batcher, ModuleBase]:
        """Perform Batcher wrapping on the given module if configured, otherwise
        return the model as is
        """
        batch_config = get_config().runtime.batching.get(
            model_type,
            get_config().runtime.batching.get("default", {}),
        )
        log.debug2("Batch config for model type [%s]: %s", model_type, batch_config)
        batch_size = batch_config.get("size", 0)
        if batch_size > 0:
            log.info(
                "<RUN89713768I>",
                "Enabling batch size [%s] for [%s] of type [%s]",
                batch_size,
                model_id,
                model_type,
            )
            return Batcher(
                model_name=model_id,
                model=caikit_core_model,
                batch_size=batch_size,
                batch_collect_delay_s=batch_config.get("collect_delay_s"),
            )
        return caikit_core_model



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Subprocess only tests - includes global train servicer and model train 
servicer tests"""

# Standard
import json
import uuid

# Third Party
import grpc
import pytest

# Local
from .test_model_train_servicer_impl import sample_model_train_servicer
from caikit.core import MODEL_MANAGER
from caikit.runtime.protobufs import process_pb2
from caikit.runtime.service_factory import get_train_params, get_train_request
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from sample_lib.data_model.sample import SampleTrainingType
from sample_lib.modules import SampleModule
from tests.conftest import random_test_id, set_use_subprocess
from tests.fixtures import Fixtures
import caikit.core
import sample_lib


@pytest.mark.parametrize("oom_exit_code", [137, 9])
def test_global_train_returns_exit_code_with_oom(
    sample_train_servicer, reset_model_manager, oom_exit_code
):
    """Test that if module goes into OOM we are able to surface error code"""
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(
        jsondata=stream_type.JsonData(data=[SampleTrainingType(1)])
    )
    train_class = get_train_request(SampleModule)
    train_request_params_class = get_train_params(SampleModule)
    train_request = train_class(
        model_name=random_test_id(),
        parameters=train_request_params_class(
            batch_size=42,
            training_data=training_data,
            oom_exit=True,
            oom_exit_code=oom_exit_code,
        ),
    ).to_proto()

    with set_use_subprocess(True):
        training_response = sample_train_servicer.Train(
            train_request, Fixtures.build_context("foo")
        )
        MODEL_MANAGER.get_model_future(training_response.training_id).wait()

        future_info = MODEL_MANAGER.get_model_future(
            training_response.training_id
        ).get_info()
        assert f"Training process died with OOM error!" in str(future_info.errors[0])


@pytest.mark.parametrize("oom_exit_code", [137, 9])
def test_model_train_memory_error_raises(
    sample_model_train_servicer, reset_model_manager, oom_exit_code
):
    """Test that if there is memory error we are able to raise it"""
    with pytest.raises(CaikitRuntimeException) as e:
        context = Fixtures.build_context("foo")

        training_id = str(uuid.uuid4())
        model_train_request = process_pb2.ProcessRequest(
            trainingID=training_id,
            request_dict={
                "train_module": "00110203-0405-0607-0809-0a0b02dd0e0f",
                "training_params": json.dumps(
                    {
                        "model_name": "abc",
                        "parameters": {
                            "training_data": {
                                "jsondata": {
                                    "data": [
                                        sample_lib.data_model.SampleTrainingType(
                                            number=1
                                        ).to_dict()
                                    ]
                                },
                            },
                            "oom_exit": True,
                            "oom_exit_code": oom_exit_code,
                        },
                    }
                ),
            },
            training_input_dir="training_input_dir",
        )
        with set_use_subprocess(True):
            sample_model_train_servicer.Run(model_train_request, context)

    assert "OOM error during training" in str(e.value)
    assert e.value.status_code == grpc.StatusCode.INTERNAL



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""These are mostly some older tests that we wrote ad-hoc while throwing the inference proto-generation together

Coverage is probably not the best
"""
# Standard
from typing import List, Optional, Union
from unittest.mock import patch

# Third Party
from docstring_parser import ParseError

# Local
from caikit.core.signature_parsing.docstrings import (
    _extract_nested_type,
    _get_docstring_type,
    get_arg_type,
    get_return_type,
    is_optional,
)
from sample_lib.data_model import SampleInputType
import caikit
import sample_lib


def test_get_docstring_type():
    # TODO: fun edge case where producer word in description is found in types
    # works on a sample_lib type
    assert (
        _get_docstring_type(
            candidate_type_names=["sample_lib.data_model.SampleOutputType"],
        )
        == sample_lib.data_model.SampleOutputType
    )

    # works on nested type
    assert (
        _get_docstring_type(
            candidate_type_names=["List(int)"],
        )
        == List[int]
    )

    # works on a well known type
    assert (
        _get_docstring_type(
            candidate_type_names=["ProducerId"],
        )
        == caikit.core.data_model.ProducerId
    )

    # returns a Union on valid types
    assert (
        _get_docstring_type(
            candidate_type_names=[
                "SampleInputType",
                "sample_lib.data_model.SampleOutputType",
            ],
        )
        == Union[
            sample_lib.data_model.SampleOutputType,
            sample_lib.data_model.SampleInputType,
        ]
    )


def test_extract_nested_type():
    """
    Test that this function returns the type of List[T], or None if it's not a nested type
    """
    assert _extract_nested_type("List[str]") == List[str]
    assert _extract_nested_type("list(str)") == List[str]
    assert _extract_nested_type("List(SampleInputType)") == List[SampleInputType]
    assert _extract_nested_type("int") == None


def test_is_optional_works_on_corner_cases_docstrings():
    """
    Test that is_optional works on different corner cases of docstrings for an optional field
    """
    # test docstring with multiple params of the same name still works, but
    # evaluated whether "optional" on the first description
    def _fn(self, some_input: Optional[int]):
        """
        Args:
            some_input: yadda yadda blah int
                optional int for input
            some_input: I repeat myself

        Returns:
            None
        """
        pass

    assert is_optional(_fn, "some_input") is True

    # test if docstring_parser.parse throws an exception, we return False
    with patch("docstring_parser.parse", side_effect=ParseError("mocked error")):
        assert is_optional(_fn, "some_input") is False

    def _fn(input):
        pass

    # test is_optional works on empty docstring
    assert is_optional(_fn, "input") is False


def test_get_arg_type_works_on_corner_cases_docstrings():
    """
    Test that get_arg_type works on different corner cases of docstrings
    """
    # test docstring with multiple params of the same name still works, but
    # evaluated the type on the first description
    def _fn(self, some_input: str):
        """

        Args:
            some_input: yadda yadda blah int
                int for input
            some_input: I repeat myself but type str here
                str for input

        Returns:
            None
        """
        pass

    assert get_arg_type(_fn, "some_input") == int

    # test if docstring_parser.parse throws an exception, we return None
    with patch("docstring_parser.parse", side_effect=ParseError("mocked error")):
        assert get_arg_type(_fn, "some_input") is None

    # test get_arg_type works on empty docstring
    def _fn_no_docstring(input):
        pass

    assert get_arg_type(_fn_no_docstring, "input") == None


def test_get_return_type_corner_case_with_exception():
    def _fn():
        pass

    # test get_return_type works on empty docstring
    assert get_return_type(_fn) == None

    # test if docstring_parser.parse throws an exception, we swallow and return none
    with patch("docstring_parser.parse", side_effect=ParseError("mocked error")):
        assert get_return_type(_fn) is None



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Local
from caikit.core.data_model import DataStream
from caikit.core.data_model.streams.validator import DataStreamValidator
from caikit.core.exceptions.validation_error import DataValidationError

# Unit Test Infrastructure
from tests.base import TestCaseBase

LIST_OF_DICTS = [
    {"a": 1, "b": 2, "c": 3},
    {"a": 4, "b": 5, "c": 6},
    {"a": 7, "b": 8, "c": 9},
]

LIST_OF_LISTS = [
    [1, 3],
    [4, 6],
    [7, 9],
]


class TestDataStreamValidator(TestCaseBase):
    def test_it_throws_validation_error_on_missing_dictionary_keys(self):
        dict_stream = DataStream.from_iterable(LIST_OF_DICTS)
        # `d` does not exist on the input stream
        validator = DataStreamValidator({"a": int, "b": int, "c": int, "d": int})
        dict_stream = validator.validate(dict_stream)
        # Since validation is lazy, we will not error until we evaluate the stream
        with self.assertRaises(DataValidationError):
            list(dict_stream)

    def test_it_throws_validation_error_on_input_lists_that_are_too_small(self):
        list_stream = DataStream.from_iterable(LIST_OF_LISTS)
        # input lists have only 2 elements each
        validator = DataStreamValidator({"a": int, "b": int, "c": int})
        list_stream = validator.validate(list_stream)
        with self.assertRaises(DataValidationError):
            list(list_stream)

    def test_it_throws_validation_error_on_input_lists_that_are_too_long(self):
        list_stream = DataStream.from_iterable(LIST_OF_LISTS)
        # input lists have 2 elements each
        validator = DataStreamValidator({"a": int})
        list_stream = validator.validate(list_stream)
        with self.assertRaises(DataValidationError):
            list(list_stream)

    def test_it_throws_validation_error_if_types_are_wrong(self):
        list_stream = DataStream.from_iterable(LIST_OF_LISTS)
        validator = DataStreamValidator({"a": int, "b": list})
        list_stream = validator.validate(list_stream)
        with self.assertRaises(DataValidationError):
            list(list_stream)

        dict_stream = DataStream.from_iterable(LIST_OF_DICTS)
        validator = DataStreamValidator({"a": int, "b": list, "c": str})
        dict_stream = validator.validate(dict_stream)
        with self.assertRaises(DataValidationError):
            list(dict_stream)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for producer"""

# Local
from caikit.core.data_model.producer import ProducerId


def test_add_producer_ids():
    prod_id_1 = ProducerId("foo", "1.2.3")
    prod_id_2 = ProducerId("bar", "1.2.3")
    producer = prod_id_1 + prod_id_2
    assert producer.name == "foo & bar"


def test_proto_roundtrip():
    prod_id = ProducerId("foo", "1.2.3")
    prod_id_roundtrip = ProducerId.from_proto(prod_id.to_proto())
    assert prod_id.name == prod_id_roundtrip.name
    assert prod_id.version == prod_id_roundtrip.version



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Data streams for lazily loading, munging and passing data through multiple modules.
"""
# Standard
from collections.abc import Iterable
from glob import glob
from io import UnsupportedOperation
from typing import Dict, Generic, List, Tuple, TypeVar, Union
import collections
import csv
import io
import itertools
import json
import os
import random
import typing

# Third Party
import ijson

# First Party
import alog

# Local
from ...augmentors import AugmentorBase
from ...exceptions import error_handler
from ...toolkit import fileio
from ..json_dict import JsonDictValue
from .multipart_decoder import is_multipart_file, stream_multipart_file

log = alog.use_channel("DATSTRM")
error = error_handler.get(log)

T = TypeVar("T")

# ghart: These public methods are all needed. This class is essentially its own factory, so these
# are all the different ways of coercing different data sources into a common stream class
# pylint: disable=too-many-public-methods
class DataStream(Generic[T]):
    """A data stream is a iterable container class that is reentrant in the sense that it can be
    iterated over multiple times.  The items produced by a data stream may be any python object
    and are called data items.  The data items produced by an iterator over a data stream are
    generated lazily (unless the `.eager` method is called) so that each data item in a series of
    data streams is produced as it is accessed.  This allows processing datasets that are too large
    to fit into memory.  A number of functional style methods are provided for manipulating and
    munging data streams and the `.stream` method on modules can also be used to
    process data streams.

    The `DataStream` class is really just a generic wrapper around functions that produce python
    iterators or generators.
    """

    def __init__(self, generator_func, *args, **kwargs):
        """Create a new `DataStream` from a function that creates a python generator or iterator
        over the desired data items.

        Args:
            generator_func (callable(*args, **kwargs)): A function that, when
                called, either (a) constructs a generator or (b) returns a
                python iterator that yields data items, which may be any python
                or data model object.  Each time `generator_func` is called, it
                must recreate the same generator/iterator.  `generator_func`
                must also produce its elements lazily. If `generator_func`
                returns, say a list or tuple, then all of the data will be
                loaded into memory immediately instead of lazily.
                `generator_func` is invoked every time that a `DataStream` is
                iterated over, i.e., when `__iter__` is called.

            args, kwargs: Additional arguments passed to `generator_func`.  These are passed every
                time that `generator_func` is called, i.e., every time we iterate over the data
                stream. These arguments are generally useful for passing arguments to an initial
                data loader function (see `.from_csv` for an example).  In order to retain other
                variables in a `generator_func` consider relying on closures instead of arguments.

        Notes:
            The constructor of `DataStream` is not usually invoked directly.  The typical use case
            is to construct a data stream using one of the `.from_` class methods or else from the
            `.stream` method of a module or by extending the `DataStream` class.

            Lexical closures, generators and iterators are all important to understand when writing
            a new `generator_func`.  Consider reviewing these topics before writing custom generator
            functions.
        """
        if not callable(generator_func):
            error(
                "<COR83886275E>",
                TypeError("Data stream `generator_func` is not callable."),
            )

        self.generator_func = generator_func
        self.generator_args, self.generator_kwargs = args, kwargs
        self._length = None

    @classmethod
    def from_iterable(cls, data: typing.Iterable[T]) -> "DataStream[T]":
        """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            data (iterable): A list or tuple or other python iterable used to
                construct a new data stream where each data item contains a
                single data item.

        Returns:
            DataStream: A new data stream that produces data items from the
                elements of `data`.

        Examples:
            >>> list_stream = DataStream.from_iterable([1, 2, 3])
            >>> for data_item in list_stream:
            >>>     print(data_item)
            1
            2
            3
        """
        error.type_check("<COR88684982E>", Iterable, data=data)
        return cls(cls._from_iterable_generator, data)

    @classmethod
    def _from_iterable_generator(cls, data: typing.Iterable[T]) -> typing.Iterator[T]:
        return iter(data)

    @classmethod
    def from_jsonl(cls, filename: str) -> "DataStream[Dict]":
        """Creates a new data stream from a path to a file with JSON lines array, where
        each line is a valid JSON (python dict)

        Args:
            filename (str): A path to a utf8 encode text file with JSON lines
                array, where each line is a valid JSON (python dict)

        Returns:
            DataStream: A new data stream that produces python dict items each
                containing a single JSON object corresponding to each line

        Notes:
            This class method returns a data stream over the valid JSON objects and each
            JSON object is on one line.

            https://jsonlines.org/

        Examples:
            For a JSON lines file that looks like:
                {"name": "Gilbert", "wins": [["straight", "7♣"], ["one pair", "10♥"]]}
                {"name": "Alexa", "wins": [["two pair", "4♠"], ["two pair", "9♠"]]}
                {"name": "May", "wins": []}
                {"name": "Deloise", "wins": [["three of a kind", "5♣"]]}

            >>> jsonl_data_stream = DataStream.from_jsonl('sample.jsonl')
            >>> for data_item in jsonl_data_stream:
            >>>     print(data_item)
            {'name': 'Gilbert', 'wins': [['straight', '7♣'], ['one pair', '10♥']]}
            {'name': 'Alexa', 'wins': [['two pair', '4♠'], ['two pair', '9♠']]}
            {'name': 'May', 'wins': []}
            {'name': 'Deloise', 'wins': [['three of a kind', '5♣']]}

        """
        error.file_check("<COR32600575E>", filename)

        return cls(cls._from_jsonl_generator, filename)

    @classmethod
    def _from_jsonl_generator(cls, filename):
        with open(filename, mode="rb") as json_fh:
            log.debug2("Loading JSON array file:  %s", filename)
            lines = json_fh.readlines()

            try:
                for line in lines:
                    if line.strip():  # ignore empty lines
                        yield json.loads(line)
            except json.JSONDecodeError as e:
                error(
                    "<COR55596551E>",
                    ValueError(f"Invalid JSON object in `{line}`, error: {e.msg}"),
                )
            except TypeError:
                error(
                    "<COR35596551E>",
                    ValueError("Invalid JSON object in `{}`".format(line)),
                )

    @classmethod
    def from_json_array(cls, filename: str) -> "DataStream[Dict]":
        """Creates a new data stream from a path to a file with JSON array, where each item is a
        valid JSON (python dict)

        Args:
            filename (str): A path to a utf8 encode text file with JSON array,
                where each item is a valid JSON (python dict)

        Returns:
            DataStream: A new data stream that produces python dict items each
                containing a single JSON object specified by 'filename'

        Notes:
            This class method returns a data stream over the valid JSON objects of a single
            JSON array text file.

        Examples:
            For a JSON file that looks like:
                [
                { a: 1, b: 2, c: False },
                { a: 2, b: 3 },
                { a: 3, c: True }
                ]

            >>> json_data_stream = DataStream.from_json_array('sample.json')
            >>> for data_item in json_data_stream:
            >>>     print(data_item)
            { a: 1, b: 2, c: False }
            { a: 2, b: 3 }
            { a: 3, c: True }

        """
        error.file_check("<COR39609575E>", filename)

        return cls(cls._from_json_array_file_generator, filename)

    @classmethod
    def _from_json_array_file_generator(cls, filename):
        # open the file
        with open(filename, mode="rb") as json_fh:
            log.debug2("Loading JSON array file: %s", filename)
            yield from cls._from_json_array_buffer_generator(json_fh, filename)

    @classmethod
    def _from_json_array_buffer_generator(cls, json_fh: typing.IO, filename: str = ""):
        try:
            # For re-entrance
            json_fh.seek(0)
        except UnsupportedOperation:
            error(
                "<COR59442457E>",
                RuntimeError(
                    "File handler for json array in filename {} not seekable".format(
                        filename
                    )
                ),
            )
        # For each {} object of the array
        try:
            item_idx = None
            for item_idx, obj in enumerate(ijson.items(json_fh, "item")):
                log.debug2("Loading object index %d", item_idx)
                yield obj
            if item_idx is None:
                # Not an array
                error(
                    "<COR79428339E>",
                    ValueError("Non-array JSON object in `{}`".format(filename)),
                )
        except ijson.JSONError:
            error(
                "<COR85596551E>",
                ValueError("Invalid JSON object in `{}`".format(filename)),
            )

    @classmethod
    def from_csv(cls, filename: str, *args, skip=0, **kwargs) -> "DataStream[List]":
        """Create a new data stream from a csv (comma separated value) file where each data item
        corresponds to a line of the csv file and consists of a list containing the comma separated
        values.

        Args:
            filename (str): A path to a csv file that has rows corresponding to
                data items and columns corresponding to the elements of each
                data item.
            skip (int): Number of lines to skip at the beginning of the csv
                file.  This is often useful for skipping a header line.
            args, kwargs: Additional arguments passed to the `csv.reader` function.
                These can be used to specify the delimiter or other csv settings.

        Returns:
            DataStream: A data stream that produces a data item for each line of
                the csv file and where each element of the data item corresponds
                to a column in the csv file.Examples:
            For a sample.csv that looks like:
                a, b, c
                d, e, f
            >>> csv_stream = DataStream.from_csv('sample.csv')
            >>> for data_item in csv_stream:
            >>>     print(data_item)
            ['a', 'b', 'c']
            ['d', 'e', 'f']
        """
        # verify that the csv file exists and is a regular file
        if not os.path.exists(filename) or not os.path.isfile(filename):
            error(
                "<COR82308234E>",
                FileNotFoundError(
                    "csv filename `{}` does not exist or is not a regular file.".format(
                        filename
                    )
                ),
            )

        return cls(cls._from_csv_generator, filename, skip, *args, **kwargs)

    @classmethod
    def _from_csv_generator(cls, filename, skip, *csv_args, **csv_kwargs):
        # open the csv file (closure around `filename`)
        with open(filename, mode="r", encoding="utf8") as fh:
            # skip lines if requested
            for _ in range(skip):
                # pylint: disable=stop-iteration-return
                next(fh)

            # for each line of the csv file, yield a list
            for line in csv.reader(fh, *csv_args, **csv_kwargs):
                yield line

    @classmethod
    def from_header_csv(cls, filename: str, *args, **kwargs) -> "DataStream[Dict]":
        """Create a new data stream from a csv where the first row is a header
        and each subsequent row is an element. The yielded elements are tuples
        of dicts where each dict pairs the row values with the corresponding
        column headers.

        Args:
            filename (str): A path to a csv file that has rows corresponding to
                data items and columns corresponding to the elements of each
                data item.
            args, kwargs: Additional arguments passed to the `csv.reader` function.
                These can be used to specify the delimiter or other csv settings.

        Returns:
            DataStream: A data stream that produces a data item for each line of
                the csv file and where each element of the stream is a dict
                representation of the fieldsExamples:
            For a sample.csv that looks like:
                foo, bar, baz
                a, b, c
                d, e, f
            >>> csv_stream = DataStream.from_csv('sample.csv')
            >>> for data_item in csv_stream:
            >>>     print(data_item)
            {"foo": "a", "bar": "b", "baz": "c"}
            {"foo": "d", "bar": "e", "baz": "f"}
        """
        # verify that the csv file exists and is a regular file
        if not os.path.exists(filename) or not os.path.isfile(filename):
            error(
                "<COR44308234E>",
                FileNotFoundError(
                    "csv filename `{}` does not exist or is not a regular file.".format(
                        filename
                    )
                ),
            )

        return cls(cls._from_header_csv_generator, filename, *args, **kwargs)

    @classmethod
    def _from_header_csv_generator(cls, filename, *csv_args, **csv_kwargs):
        # open the csv file (closure around `filename`)
        with open(filename, mode="r", encoding="utf8") as fh:
            yield from cls._from_header_csv_buffer_generator(
                fh, *csv_args, **csv_kwargs
            )

    @classmethod
    def _from_header_csv_buffer_generator(cls, fh: typing.IO, *csv_args, **csv_kwargs):
        try:
            # For re-entrance
            fh.seek(0)
        except UnsupportedOperation:
            error(
                "<COR55834083E>",
                RuntimeError("File handler for csv with header not seekable"),
            )
        # for each line of the csv file, yield a dict
        for line in csv.DictReader(fh, *csv_args, **csv_kwargs):
            yield line

    @classmethod
    def from_txt(cls, filename: str) -> "DataStream[str]":
        """Create a new data stream from a path to a utf8 encoded text file where each data item
        corresponds to a single line of the file.

        Args:
            filename (str): A path to a utf8 encode text file with each line
                corresponding to a data item.

        Returns:
            DataStream: A new data stream that produces string data items each
                containing a single line from the file specified by `filename`.

        Notes:
            This class method returns a data stream over the lines of a single text file.  In
            order to construct a datastream over separate files, rather than lines, consider using
            `.from_txt_collection`.

        Examples:
            For a text file that looks like:
                first line
                second line
                third line

            >>> txt_line_stream = DataStream.from_file('sample.txt')
            >>> for data_item in txt_line_stream:
            >>>     print(data_item)
            first line
            second line
            third line
        """
        error.file_check("<COR79693043E>", filename)

        return cls(cls._from_txt_generator, filename)

    @classmethod
    def _from_txt_generator(cls, filename):
        # open the file (closure around `filename`)
        with open(filename, mode="r", encoding="utf8") as fh:
            # for each line of the file
            for line in fh:
                # strip new lines and carriage returns and yield the line
                yield line.rstrip("\n\r")

    @classmethod
    def from_file(cls, filename: str) -> "DataStream[Union[Dict, Tuple, str]]":
        """Loads up a DataStream from a file. Will call the correct DataStream.from_*
        static constructor based on the file extension

        The data items returned in the data stream are:
        For JSON:
            dictionaries
        For all other files (besides CSV for now)
            strings (1 per line)

        Args:
            filename (str): Name of file

        Returns:
            DataStream: Resulting datastream from file
        """
        # file detection
        _, file_ext = os.path.splitext(filename)
        # choose the right from_* fn
        if file_ext.lower() == ".json":
            log.debug2("Detected .json extension, loading %s as a json file", filename)
            return DataStream.from_json_array(filename)

        if file_ext.lower() == ".csv":
            log.debug2("Detected .csv extension, loading %s as a csv file", filename)
            return DataStream.from_csv(filename)

        log.debug2("Loading %s as a raw text file", filename)
        # TODO: test this at some point (this path is unused currently)
        return DataStream.from_txt(filename)

    @classmethod
    def _from_collection(
        cls, dirname: str, extension: str, file_opener
    ) -> "DataStream[Union[Dict, Tuple, str]]":
        """Create a new data stream from a path containing multiple files where
        each data item corresponds to the entire serialized content in a single file. The
        file_handler function does the serialization of individual files

        Args:
            dirname (str): A directory path containing a number of utf8 encoded
                text files with the `.txt` filename extension.
            extension (str): Extension of the file. Note that all files are read
                in the same utf8 encoding.
            file_opener (function): Function to deserialize a file on disk to
                memory

        Returns:
            DataStream: A new data stream that produces string data items each
                containing the text contained in a single file found in
                `dirname`.

        Notes:
            Each data item in this data stream represents the *entire* text contained in a single
            file and are not split by line or otherwise.
        """
        # verify that `dirname` exists
        cls._verify_dir(dirname)

        return cls(cls._from_collection_generator, dirname, extension, file_opener)

    @classmethod
    def _from_collection_generator(cls, dirname, extension, file_opener):
        # glob `*.txt` files in `dirname` (closure around `dirname`)
        for filename in glob(os.path.join(dirname, "*." + extension)):
            yield file_opener(filename)

    @classmethod
    def from_txt_collection(cls, dirname: str, extension="txt") -> "DataStream[str]":
        """Create a new data stream from a path containing multiple utf8 encoded text files where
        each data item corresponds to the entire text contained in a single file.

        Args:
            dirname (str): A directory path containing a number of utf8 encoded
                text files with the `.txt` filename extension.
            extension: str (Optional)
                Optional extension of the text file. Note that all files are read in the same
                utf8 encoding. Defaults to 'txt'

        Returns:
            DataStream: A new data stream that produces string data items each
                containing the text contained in a single `.txt` (or specified
                extension) file found in `dirname`.

        Notes:
            Each data item in this data stream represents the *entire* text contained in a single
            file and are not split by line or otherwise.
        """
        return cls._from_collection(dirname, extension, fileio.load_txt)

    @classmethod
    def from_json_collection(
        cls, dirname: str, extension="json"
    ) -> "DataStream[Union[Dict, Tuple, List]]":
        """Create a new data stream from a path containing multiple JSON files where
        each data item corresponds to the entire serialized JSON contained in a single file.

        Args:
            dirname (str): A directory path containing a number of utf8 encoded
                text files with the `.txt` filename extension.
            extension: str (Optional)
                Optional extension of the JSON file. Note that all files are read in the same
                utf8 encoding. Defaults to 'json'

        Returns:
            DataStream: A new data stream that produces string data items each
                containing the text contained in a single `.json` (or specified
                extension) file found in `dirname`.

        Notes:
            Each data item in this data stream represents the *entire* text contained in a single
            file and are not split by line or otherwise.
        """
        return cls._from_collection(dirname, extension, fileio.load_json)

    @classmethod
    def from_csv_collection(cls, dirname: str) -> "DataStream[Dict]":
        """Create a new data stream by chaining data streams from each of the file from a path
        containing multiple csv files where each file can have 1 or more data item.

        Args:
            dirname (str): A directory path containing a number of csv files

        Returns:
            DataStream: A new data stream that is chained from all data streams
                by reading (from_header_csv) all files in all `.csv` files found
                in `dirname`. All data items are dicts.
        """
        # verify that `dirname` exists
        cls._verify_dir(dirname)

        return cls(cls._from_csv_collection_generator, dirname)

    @classmethod
    def _from_csv_collection_generator(cls, dirname):
        # list of data_streams created from different files
        data_stream_list = []
        # glob `*.txt` files in `dirname` (closure around `dirname`)
        for filename in glob(os.path.join(dirname, "*.csv")):
            data_stream_list.append(cls.from_header_csv(filename=filename))
        # yield the combined data item once flattened
        for data_item in DataStream.chain(data_stream_list).flatten():
            yield data_item

    @classmethod
    def from_jsonl_collection(cls, dirname: str) -> "DataStream[Dict]":
        """Create a new data stream by chaining data streams from each of the file from a path
        containing multiple jsonl files where each file can have 1 or more data item.

        Args:
            dirname (str): A directory path containing a number of jsonl files

        Returns:
            DataStream: A new data stream that is chained from all data streams
                by reading (from_jsonl) all files in all `.jsonl` files found in
                `dirname`.
        """
        # verify that `dirname` exists
        cls._verify_dir(dirname)

        return cls(cls._from_jsonl_collection_generator, dirname)

    @classmethod
    def _from_jsonl_collection_generator(cls, dirname):
        # list of data_streams created from different files
        data_stream_list = []
        # glob `*.txt` files in `dirname` (closure around `dirname`)
        for filename in glob(os.path.join(dirname, "*.jsonl")):
            data_stream_list.append(cls.from_jsonl(filename=filename))
        # yield the combined data item once flattened
        for data_item in DataStream.chain(data_stream_list).flatten():
            yield data_item

    @classmethod
    def from_multipart_file(cls, filename: str) -> "DataStream[JsonDictValue]":
        """Loads up a DataStream from a multipart file

        The data items returned in the data stream are determined by the
        content type for each part in the multipart file by calling
        the correct DataStream.from_*

        Args:
            filename (str): Name of file

        Returns:
            DataStream: Resulting datastream from file
        """
        error.value_check(
            "<COR04987251E>", is_multipart_file(filename), "file is not multipart"
        )
        stream_list = []
        for part in stream_multipart_file(filename):
            content_type = part.content_type
            if "json" in content_type:
                stream_list.append(
                    cls(cls._from_json_array_buffer_generator, part.fp, part.filename)
                )
            elif "csv" in content_type:
                stream_list.append(
                    cls(
                        cls._from_header_csv_buffer_generator,
                        _UtfEncodeIOWrapper(part.fp),
                    )
                )
            else:
                error(
                    "<COR91833046E>",
                    ValueError("Unsupported content type: {}".format(content_type)),
                )
        return DataStream.chain(*stream_list)

    def train_test_split(
        self, test_split=0.25, seed=None
    ) -> "Tuple[DataStream[T], DataStream[T]]":
        """Split the current datastream into train/test substreams.

        Args:
            test_split (float): The fraction of examples to assign to the test
                substream, in [0, 1]
            seed (int | None): The seed for initializing the random assignment.
                If not provided, a randomly chosen seed will be used.

        Returns:
            tuple(DataStream, DataStream): Two substreams: a train set
                substream, and a test set substream
        """
        assert 0.0 <= test_split <= 1.0

        if seed is None:
            seed = random.randint(0, 10000)

        def train_generator_func():
            rng = random.Random(seed)
            for data_item in self:
                if rng.random() > test_split:
                    yield data_item

        def test_generator_func():
            rng = random.Random(seed)
            for data_item in self:
                if rng.random() <= test_split:
                    yield data_item

        return DataStream(train_generator_func), DataStream(test_generator_func)

    # pylint: disable=no-self-argument
    def chain(*args) -> "DataStream":
        """Chain multiple data streams together sequentially.  The returned data stream produces
        the data items from each passed data stream in turn.

        Args:
            args (tuple(DataStream)): A tuple containing the data streams to
                chain, passed as variadic arguments.

        Returns:
            DataStream: A new data stream that produces the data items from the
                provided data streams sequentially.
        """
        return DataStream(lambda: itertools.chain(*args))

    # pylint: disable=keyword-arg-before-vararg
    def filter(
        self, func=lambda data_item: data_item, *args, **kwargs
    ) -> "DataStream[T]":
        """Skip elements in the data stream as identified by a passed function.

        Args:
            func (callable(data_item)): The function used to identify data items
                that will be filtered.  The function takes a single data item as
                an argument and returns `True` in order to keep the element and
                `False` in order to skip it.  The default filter function
                removes falsey values.

        Returns:
            DataStream: A new data stream that produces the data items from the
                current data stream only when `func` evaluates to true.
        """
        error.value_check(
            "<COR59884427E>", callable(func), "filter function is not callable"
        )

        return DataStream(
            lambda: (
                data_item for data_item in self if func(data_item, *args, **kwargs)
            )
        )

    def shuffle(self, buffer_size, seed=None) -> "DataStream[T]":
        """Randomly shuffles the elements of this dataset. If buffer_size is smaller than the full
        size of the full data stream, it is a partial random shuffle which is similar to
        Tensorflow's dataset shuffle. For instance, if your dataset contains 10,000 elements but
        buffer_size is set to 1,000, then shuffle will initially select a random element from only
        the first 1,000 elements in the buffer. Once an element is selected, its space in the
        buffer is replaced by the next (i.e. 1,001-st) element, maintaining the 1,000 element
        buffer.

        Args:
            buffer_size (int): the size of the buffer space, should be greater
                than 0
            seed (int | None): The seed for initializing the random assignment.
                If not provided, a randomly chosen seed will be used.

        Returns:
            DataStream: A new data stream after shuffled.
        """
        # make sure buffer space is valid
        error.type_check("<COR06395206E>", int, buffer_size=buffer_size)
        error.value_check(
            "<COR78471251E>", buffer_size > 0, "Buffer size must be an int > 0"
        )

        if seed is None:
            seed = random.randint(0, 10000)

        def generator_func():
            buffer = []
            random.seed(seed)
            if self._length is not None and self._length <= buffer_size:
                buffer = list(itertools.islice(self, buffer_size))
            else:
                count = 0
                for e in self:
                    if count < buffer_size:
                        buffer.append(e)
                        count += 1
                    else:
                        idx = random.randint(0, buffer_size - 1)
                        item = buffer[idx]
                        buffer[idx] = e
                        yield item

            random.shuffle(buffer)
            for item in buffer:
                yield item

        return DataStream(generator_func)

    def eager(self) -> "DataStream[T]":
        """Evaluate the data stream, place it into memory and return a new data stream over these
        static values.  This is useful if your data stream can fit into memory, at least up to a
        certain point, and it will not be efficient to lazily and, potentially, re-evaluate the
        stream each time it is iterated over.

        Returns:
            DataStream: A new data stream that iterates over the evaluated, in-
                memory data items in this stream.
        """
        return DataStream.from_iterable(list(self))

    def map(self, func, *args, **kwargs) -> "DataStream":
        """Apply a function to each element in the data stream.

        Args:
            func (callable(*args, **kwargs)): A function this is lazily applied
                to each element in the data stream.
            *args, **kwargs
                Additional arguments to pass `func`.

        Returns:
            DataStream: A new data stream with `func` applied to each element.
        """
        return DataStream(
            lambda: (func(data_item, *args, **kwargs) for data_item in self)
        )

    def flatten(self) -> "DataStream":
        """Convert a 2-level nested stream into a flattened stream

        Returns:
            DataStream: A new data stream with inner stream items 'flattened'
        """

        def generator_func():
            for inner_stream in self:
                for data_item in inner_stream:
                    yield data_item

        return DataStream(generator_func)

    # pylint: disable=no-self-argument
    def zip(*args) -> "DataStream":
        """Combine the data items of multiple data streams together in tuples.

        Args:
            args (tuple(DataStream)): A tuple containing the data streams to be
                zip, passed as variadic arguments.

        Returns:
            DataStream: A data stream that produces the zipped data items.

        Notes:
            A `ValueError` is raised when the stream is iterated over if any of the zipped data
            streams do not have the same length.  Since streams are evaluated lazily, however, this
            error condition will only be detected and raised when the stream is being iterated over.
        """
        error.type_check_all("<COR19533030E>", DataStream, args=args)

        def generator_func():
            # create a unique object as a sentinel
            sentinel = object()

            # zip the data items together and pad with the sentinel
            for zipped_data_items in itertools.zip_longest(*args, fillvalue=sentinel):
                # if the sentinel is detected, the data streams do not have the same length
                if sentinel in zipped_data_items:
                    error(
                        "<COR83794589E>",
                        ValueError(
                            "Failed to zip data streams with different lengths."
                        ),
                    )

                yield zipped_data_items

        return DataStream(generator_func)

    def peek(self) -> T:
        """Returns the first element of the stream, or raises IndexError if stream is empty"""
        try:
            return next(iter(self))
        except StopIteration:
            error.log_raise("<COR48484123E>", IndexError("Cannot peek empty stream"))

    def augment(
        self,
        augmentor,
        aug_cycles,
        *,
        post_augment_func=None,
        augment_index=None,
        enforce_determinism=True,
    ) -> "DataStream[T]":
        error.type_check("<COR45851623E>", int, aug_cycles=aug_cycles)
        error.type_check("<COR80001982E>", AugmentorBase, augmentor=augmentor)
        error.type_check(
            "<COR87701982E>", bool, enforce_determinism=enforce_determinism
        )
        error.value_check(
            "<COR56795914E>", aug_cycles > 0, "Augmentation cycles must be an int > 0"
        )
        if post_augment_func is not None and not callable(post_augment_func):
            error(
                "<COR32996115E>",
                TypeError("Post augmentation operation is not callable"),
            )
        contains_iterables = isinstance(self.peek(), (tuple, list))
        # Explode if we have an augment index that is not applicable (i.e., don't have lists or
        # tuple objects being considered to apply it against).
        if not contains_iterables and augment_index is not None:
            error(
                "<COR31116115E>",
                ValueError(
                    "augment_index cannot be used unless stream contains lists/tuples"
                ),
            )
        # Explode if we don't have an augmentation index, but we need one (i.e., we have lists or
        # tuple objects, but don't know how to use the augmentor)
        if contains_iterables and augment_index is None:
            error(
                "<COR31316445E>",
                ValueError(
                    "augment_index must be given to augment a stream of lists/tuples"
                ),
            )

        def generator_func():
            for cycle_num in range(aug_cycles + 1):
                for obj in self:
                    if (
                        contains_iterables
                        and augment_index
                        and len(obj) <= augment_index
                    ):
                        error(
                            "<COR31352545E>",
                            IndexError(
                                "augment_index is out of bounds of obj in stream"
                            ),
                        )

                    # Figure out what we need to apply the augmentor to, then apply it. After
                    # that, apply the post augmentation func if one is provided.
                    augmentable = obj if augment_index is None else obj[augment_index]
                    # If this is the first cycle, don't apply the augmentor, otherwise do it.
                    augmented = (
                        augmentor.augment(augmentable) if cycle_num else augmentable
                    )
                    # In some special cases, we've designed things following the augmentor pattern
                    # that may return None for some inputs. If the augmentor is designed for this,
                    # i.e., sets .produces_none=True in the subclass, filter these objects out
                    # from the returned stream. Note that this WILL drop None values from the
                    # original dataset at the moment as well, as keeping None in the input stream
                    # is presumably a rare behavior.
                    if augmented is None and not augmentor.produces_none:
                        error(
                            "<COR34377515E>",
                            ValueError("Augmentor produced [None] unexpectedly"),
                        )
                    elif augmented is None:
                        continue

                    if post_augment_func is not None:
                        augmented = post_augment_func(augmented)
                    # If there is no augmentation index, we're done - return the augmented object
                    if augment_index is None:
                        yield augmented
                    # Otherwise we need to repack the augmented object back into the tuple,
                    # where it lives in peace with everything else in this item.
                    else:
                        yield tuple(
                            elem if idx != augment_index else augmented
                            for idx, elem in enumerate(obj)
                        )
            # Reset the augmentor after all cycles to ensure the DataStream is deterministic
            if enforce_determinism:
                augmentor.reset()

        return DataStream(generator_func)

    def __add__(self, other):
        """The addition operator for data streams is equivalent to calling `.chain` and combines
        this data stream with another sequentially.
        """
        return self.chain(other)

    def __getitem__(self, idx) -> T:
        """Index or slice each data item.  This is valuable for creating new data streams over the
        elements of a stream that produces tuples, lists, arrays, et cetra.

        Args:
            idx (int or slice): The index or slice to be applied to each data
                item.

        Returns:
            DataStream: A new data stream with `data_item[idx]` applied to each
                data item.

        Notes:
            This operation may be somewhat counter intuitive since `data_stream[0]` does not return
            the first element of the data stream and, instead, returns a new data stream that
            produces `data_item[0]` for each data item.

            This operation may fail with a `TypeError` if the data items in the stream are not
            subscriptable.
        """
        return DataStream(lambda: (data_item[idx] for data_item in self))

    def __iter__(self):
        """Return an iterator or generator over all of the data items in this data stream.  Data
        streams are reentrant in the sense that they can be iterated over multiple times.
        """
        # call the generator function to create an iterable yielding the data items
        # pass in the variadic arguments saved during construction of the data stream
        generator = self.generator_func(*self.generator_args, **self.generator_kwargs)
        if not isinstance(generator, collections.abc.Iterable):
            error(
                "<COR35849950E>",
                RuntimeError("`generator_func` did not return an iterable"),
            )

        return generator

    def __len__(self):
        """Return the number of data items contained in this data stream.  This requires that the
        data stream be iterated over, which may be time consuming.  This value is then stored
        internally so that subsequent calls do not iterate over the data stream again.
        """
        if self._length is None:
            self._length = sum(1 for data_item in self)

        return self._length

    def __or__(self, module):
        """Feed this data stream into the `.stream` method of a module.  This is syntactic sugar
        that allows modules to be chained like `data_stream | module1 | module2` rather than the
        equivalent `module2.stream(module1.stream(data_stream))`.
        """
        return module.stream(self)

    # Helper functions

    @staticmethod
    def _verify_dir(dirname):
        # verify that `dirname` exists
        if not os.path.exists(dirname):
            error(
                "<COR82306771E>",
                FileNotFoundError(
                    "Could not find collection directory `{}`".format(dirname)
                ),
            )

        # verify that `dirname` is a directory
        if not os.path.isdir(dirname):
            error(
                "<COR82306849E>",
                NotADirectoryError(
                    "collection path `{}` is not a directory".format(dirname)
                ),
            )


class _UtfEncodeIOWrapper(io.IOBase):
    """Lil' wrapper class to convert a bytes buffer to a string buffer"""

    def __init__(self, bytes_stream: typing.IO[bytes]):
        self.bytes_stream = bytes_stream

    def read(self, *args, **kwargs):
        res = self.bytes_stream.read(*args, **kwargs)
        return res.decode("utf-8")

    def readline(self, *args, **kwargs):
        res = self.bytes_stream.readline(*args, **kwargs)
        return res.decode("utf-8")

    def seek(self, *args, **kwargs):
        return self.bytes_stream.seek(*args, **kwargs)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module is responsible for creating service objects for the runtime to consume"""
# Standard
from enum import Enum
from types import ModuleType
from typing import Callable, Dict, Set, Type, Union
import dataclasses

# Third Party
import google.protobuf.descriptor
import google.protobuf.service
import grpc

# First Party
from py_to_proto.json_to_service import json_to_service
import aconfig
import alog

# Local
from caikit import get_config
from caikit.core import LocalBackend, ModuleBase, registries
from caikit.core.data_model.base import DataBase
from caikit.core.data_model.dataobject import _AUTO_GEN_PROTO_CLASSES
from caikit.core.exceptions import error_handler
from caikit.core.task import TaskBase
from caikit.interfaces.runtime.data_model import (
    TrainingInfoRequest,
    TrainingStatusResponse,
)
from caikit.runtime import service_generation
from caikit.runtime.service_generation.rpcs import CaikitRPCBase
from caikit.runtime.utils import import_util

log = alog.use_channel("SVC-FACTORY")
error = error_handler.get(log)

TRAINING_MANAGEMENT_SERVICE_NAME = "TrainingManagement"
TRAINING_MANAGEMENT_SERVICE_SPEC = {
    "service": {
        "rpcs": [
            {
                "name": "GetTrainingStatus",
                "input_type": TrainingInfoRequest.get_proto_class().DESCRIPTOR.full_name,
                "output_type": TrainingStatusResponse.get_proto_class().DESCRIPTOR.full_name,
            },
            {
                "name": "CancelTraining",
                "input_type": TrainingInfoRequest.get_proto_class().DESCRIPTOR.full_name,
                "output_type": TrainingStatusResponse.get_proto_class().DESCRIPTOR.full_name,
            },
        ]
    }
}


@dataclasses.dataclass
class ServicePackage:
    """Container with references to:
    - A service Class
    - A service Descriptor
    - A grpc servicer registration function
    - A client stub
    - A client messages module
    """

    service: Type[google.protobuf.service.Service]
    descriptor: google.protobuf.descriptor.ServiceDescriptor
    registration_function: Callable[
        [google.protobuf.service.Service, grpc.Server], None
    ]
    stub_class: Type
    messages: ModuleType
    caikit_rpcs: Dict[str, CaikitRPCBase]


class ServicePackageFactory:
    """Factory responsible for yielding the correct concrete ServicePackage implementation"""

    class ServiceType(Enum):
        INFERENCE = 1  # Inference service for the GlobalPredictServicer
        TRAINING = 2  # Training service for the GlobalTrainServicer
        TRAINING_MANAGEMENT = 3

    @classmethod
    def get_service_package(cls, service_type: ServiceType) -> ServicePackage:
        """Public factory API. Returns a service package of the requested type, from the
        configured source.

        Args:
            service_type (ServicePackageFactory.ServiceType): The type of service to build,
                to match the servicer implementation that will handle it. e.g. the
                GlobalPredictServicer expects an "INFERENCE" service

        Returns:
            ServicePackage: A container with properties referencing everything you need to bind a
                concrete Servicer implementation to a protobufs Service and grpc Server
        """
        if service_type == cls.ServiceType.TRAINING_MANAGEMENT:
            grpc_service = json_to_service(
                name=TRAINING_MANAGEMENT_SERVICE_NAME,
                package="caikit.runtime.training",
                json_service_def=TRAINING_MANAGEMENT_SERVICE_SPEC,
            )

            return ServicePackage(
                service=grpc_service.service_class,
                descriptor=grpc_service.descriptor,
                registration_function=grpc_service.registration_function,
                stub_class=grpc_service.client_stub_class,
                messages=None,  # we don't need messages here
                caikit_rpcs={},  # No caikit RPCs
            )

        # First make sure we import the data model for the correct library
        # !!!! This will use the `caikit_library` config
        _ = import_util.get_data_model()

        # Get the names for the AI domain and the proto package
        ai_domain_name = service_generation.get_ai_domain()
        package_name = service_generation.get_runtime_service_package()

        # Then do API introspection to come up with all the API definitions to support
        caikit_config = get_config()
        clean_modules = ServicePackageFactory._get_and_filter_modules(
            caikit_config, caikit_config.runtime.library
        )

        if service_type == cls.ServiceType.INFERENCE:
            rpc_list = service_generation.create_inference_rpcs(clean_modules)
            service_name = f"{ai_domain_name}Service"
        else:  # service_type == cls.ServiceType.TRAINING
            rpc_list = service_generation.create_training_rpcs(clean_modules)
            service_name = f"{ai_domain_name}TrainingService"

        rpc_list = [rpc for rpc in rpc_list if rpc.return_type is not None]

        for rpc in rpc_list:
            rpc.create_request_data_model(package_name)

        client_module = ModuleType(
            "ClientMessages",
            "Package with service message class implementations",
        )

        for proto_class in _AUTO_GEN_PROTO_CLASSES:
            # We need all the DM objects in the client_module for ease of use
            setattr(client_module, proto_class.DESCRIPTOR.name, proto_class)

        rpc_jsons = [rpc.create_rpc_json(package_name) for rpc in rpc_list]
        service_json = {"service": {"rpcs": rpc_jsons}}
        grpc_service = json_to_service(
            name=service_name, package=package_name, json_service_def=service_json
        )

        return ServicePackage(
            service=grpc_service.service_class,
            descriptor=grpc_service.descriptor,
            registration_function=grpc_service.registration_function,
            stub_class=grpc_service.client_stub_class,
            messages=client_module,
            caikit_rpcs={rpc.name: rpc for rpc in rpc_list},
        )

    # Implementation details for pure python service packages #
    @staticmethod
    def _get_and_filter_modules(
        caikit_config: aconfig.Config, lib: str
    ) -> Set[Type[ModuleBase]]:
        clean_modules = set()
        modules = [
            module_class
            for module_class in registries.module_registry().values()
            if module_class.__module__.partition(".")[0] == lib
        ]
        # NB: The `module_registry` only includes the `LOCAL` backend modules.
        # Implementations of the same module for different backends need to be fetched from the
        # backend registry
        backend_modules = []
        for backend_dict in registries.module_backend_registry().values():
            for backend, config in backend_dict.items():
                if backend != LocalBackend.backend_type:
                    backend_modules.append(config.impl_class)
        modules.extend(backend_modules)
        log.debug("Found all modules %s for library %s.", modules, lib)

        # Check config for any explicit inclusions
        included_task_types = (
            caikit_config.runtime.service_generation
            and caikit_config.runtime.service_generation.task_types
            and caikit_config.runtime.service_generation.task_types.included
        )
        included_modules = (
            caikit_config.runtime.service_generation
            and caikit_config.runtime.service_generation.module_guids
            and caikit_config.runtime.service_generation.module_guids.included
        )

        # Check config for any exclusions
        excluded_task_types = (
            caikit_config.runtime.service_generation
            and caikit_config.runtime.service_generation.task_types
            and caikit_config.runtime.service_generation.task_types.excluded
        )
        excluded_modules = (
            caikit_config.runtime.service_generation
            and caikit_config.runtime.service_generation.module_guids
            and caikit_config.runtime.service_generation.module_guids.excluded
        )

        for ck_module in modules:
            # Only create for modules from defined included and exclusion list

            if not ck_module.TASK_CLASS:
                log.debug(
                    "Skipping module %s with no task",
                    ck_module,
                )
                continue

            if (
                excluded_task_types
                and ck_module.TASK_CLASS.__name__ in excluded_task_types
            ):
                log.debug(
                    "Skipping module %s with excluded task %s",
                    ck_module,
                    ck_module.TASK_CLASS.__name__,
                )
                continue

            if excluded_modules and ck_module.MODULE_ID in excluded_modules:
                log.debug(
                    "Skipping module %s with excluded id %s",
                    ck_module,
                    ck_module.MODULE_ID,
                )
                continue

            # no inclusions specified means include everything
            if (included_task_types is None or included_task_types == []) and (
                included_modules is None or included_modules == []
            ):
                clean_modules.add(ck_module)

            # if inclusion is specified, use that
            else:
                if (included_modules and ck_module.MODULE_ID in included_modules) or (
                    included_task_types
                    and ck_module.TASK_CLASS.__name__ in included_task_types
                ):
                    clean_modules.add(ck_module)

        log.debug(
            "Filtered list of modules %s after excluding task types: %s and modules ids: %s. \
                Exclusions are defined in config",
            clean_modules,
            excluded_task_types,
            excluded_modules,
        )
        return clean_modules


def get_inference_request(
    task_or_module_class: Type[Union[ModuleBase, TaskBase]],
    input_streaming: bool = False,
    output_streaming: bool = False,
) -> Type[DataBase]:
    """Helper function to return the inference request DataModel for the Module or Task Class"""
    error.subclass_check(
        "<SVC98285724E>",
        task_or_module_class,
        ModuleBase,
        TaskBase,
    )
    task_class = (
        task_or_module_class.TASK_CLASS
        if issubclass(task_or_module_class, ModuleBase)
        else task_or_module_class
    )

    if input_streaming and output_streaming:
        request_class_name = f"BidiStreaming{task_class.__name__}Request"
    elif input_streaming:
        request_class_name = f"ClientStreaming{task_class.__name__}Request"
    elif output_streaming:
        request_class_name = f"ServerStreaming{task_class.__name__}Request"
    else:
        request_class_name = f"{task_class.__name__}Request"
    log.debug(
        "Request class name %s for class %s.", request_class_name, task_or_module_class
    )
    return DataBase.get_class_for_name(request_class_name)


def get_train_request(module_class: Type[ModuleBase]) -> Type[DataBase]:
    """Helper function to return the train request DataModel for the Module Class"""
    error.subclass_check(
        "<SVC32285724E>",
        module_class,
        ModuleBase,
    )
    request_class_name = (
        f"{module_class.TASK_CLASS.__name__}{module_class.__name__}TrainRequest"
    )
    log.debug("Request class name %s for module %s.", request_class_name, module_class)
    return DataBase.get_class_for_name(request_class_name)


def get_train_params(module_class: Type[ModuleBase]) -> Type[DataBase]:
    """Helper function to return the train parameters DataModel for the Module Class"""
    error.subclass_check(
        "<SVC98435724E>",
        module_class,
        ModuleBase,
    )
    request_class_name = (
        f"{module_class.TASK_CLASS.__name__}{module_class.__name__}TrainParameters"
    )
    log.debug("Request class name %s for module %s.", request_class_name, module_class)
    return DataBase.get_class_for_name(request_class_name)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the NLP task definitions"""
# Standard
from typing import Dict

# Third Party
import pytest

# Local
from caikit.core import ModuleBase, module
from caikit.interfaces.nlp import tasks as nlp_tasks
from tests.core.helpers import *

## Helpers #####################################################################


## Tests #######################################################################


class InvalidType:
    pass


@pytest.mark.parametrize(
    "flavor",
    (
        {"input_streaming": False, "output_streaming": False},
        {"input_streaming": False, "output_streaming": True},
    ),
)
def test_tasks(reset_globals, flavor: Dict[str, bool]):
    """Common tests for all tasks"""
    # Only support single required param named "text"
    task = nlp_tasks.TextGenerationTask
    assert set(task.get_required_parameters(flavor["input_streaming"]).keys()) == {
        "text"
    }
    input_type = task.get_required_parameters(flavor["input_streaming"])["text"]
    output_type = task.get_output_type(flavor["output_streaming"])

    # Version with the right signature and nothing else
    @module(id="foo1", name="Foo", version="0.0.0", task=task)
    class Foo1(ModuleBase):
        def run(self, text: input_type) -> output_type:
            return output_type()

    # Version with the right signature plus extra args
    @module(id="foo2", name="Foo", version="0.0.0", task=task)
    class Foo2(ModuleBase):
        def run(
            self,
            text: input_type,
            workit: bool,
            makeit: bool,
            doit: bool,
        ) -> output_type:
            return output_type()

    # Version with missing required argument
    with pytest.raises(TypeError):

        @module(id="foo3", name="Foo", version="0.0.0", task=task)
        class Foo3(ModuleBase):
            @task.taskmethod(**flavor)
            def run(self, other_name: str) -> output_type:
                return output_type()

    # Version with bad required argument type
    with pytest.raises(TypeError):

        @module(id="foo4", name="Foo", version="0.0.0", task=task)
        class Foo4(ModuleBase):
            @task.taskmethod(**flavor)
            def run(self, text: InvalidType) -> output_type:
                return output_type()

    # Version with bad return type
    with pytest.raises(TypeError):

        @module(id="foo", name="Foo", version="0.0.0", task=task)
        class Foo(ModuleBase):
            @task.taskmethod(**flavor)
            def run(self, text: input_type) -> InvalidType:
                return "hi there"



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from typing import List

# First Party
import alog

# Local
from caikit.core.data_model import (
    PACKAGE_COMMON,
    DataObjectBase,
    ProducerId,
    dataobject,
)
from caikit.core.exceptions import error_handler

log = alog.use_channel("DATAM")
error = error_handler.get(log)


@dataobject(PACKAGE_COMMON)
class ProducerPriority(DataObjectBase):
    """An ordered list of ProducerId structures in descending order of priority.
    This is used when handling conflicts between multiple producers of the same
    data structure.
    """

    producers: List[ProducerId]



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import multiprocessing
import time

# Third Party
import pytest

# Local
from caikit.core.toolkit.destroyable_process import DestroyableProcess

## Helpers #####################################################################


EXPECTED_THROW = ValueError("test-any-error")
EXPECTED_SUCCESS = "test-any-result"


def infinite_wait():
    while True:
        time.sleep(0.1)


def long_sleep():
    time.sleep(1000)


def thrower():
    raise EXPECTED_THROW


def succeeder():
    return EXPECTED_SUCCESS


@pytest.fixture(
    params=["fork", "forkserver", "spawn"],
)
def process_type(request):
    yield request.param


## Tests #######################################################################


def test_processes_can_be_interrupted(process_type):
    proc = DestroyableProcess(process_type, infinite_wait)
    proc.start()
    assert not proc.destroyed
    assert not proc.canceled
    assert not proc.ran
    assert not proc.threw
    proc.destroy()
    proc.join(60)
    assert not proc.is_alive()
    assert proc.destroyed
    assert proc.canceled
    assert proc.ran
    assert not proc.threw


def test_processes_can_return_results(process_type):
    proc = DestroyableProcess(process_type, succeeder, return_result=True)
    proc.start()
    proc.join()
    assert EXPECTED_SUCCESS == proc.get_or_throw()
    assert not proc.destroyed
    assert not proc.canceled
    assert proc.ran
    assert not proc.threw


def test_process_not_canceled_after_success(process_type):
    proc = DestroyableProcess(process_type, succeeder)
    proc.start()
    proc.join()
    assert not proc.canceled
    proc.destroy()
    assert not proc.canceled


def test_processes_can_be_set_to_not_return_results(process_type):
    proc = DestroyableProcess(process_type, succeeder, return_result=False)
    proc.start()
    proc.join()
    assert proc.get_or_throw() is None
    assert not proc.destroyed
    assert not proc.canceled
    assert proc.ran
    assert not proc.threw


def test_processes_can_throw(process_type):
    proc = DestroyableProcess(process_type, thrower)
    proc.start()
    proc.join()
    assert not proc.destroyed
    assert not proc.canceled
    assert proc.ran
    assert proc.threw

    with pytest.raises(ValueError) as ctx:
        proc.get_or_throw()

    assert str(EXPECTED_THROW) == str(ctx.value)


def test_processes_will_not_execute_if_destroyed_before_starting(process_type):
    proc = DestroyableProcess(process_type, long_sleep)
    proc.destroy()
    proc.start()
    assert not proc.is_alive()
    proc.join()
    with pytest.raises(RuntimeError):
        proc.get_or_throw()
    assert proc.destroyed
    assert proc.canceled
    assert not proc.ran
    assert proc.threw


def test_event_is_set_on_completion(process_type):
    event = multiprocessing.get_context(process_type).Event()
    proc = DestroyableProcess(process_type, succeeder, completion_event=event)
    assert not event.is_set()
    proc.start()
    proc.join()
    assert event.is_set()
    assert not proc.destroyed
    assert not proc.canceled
    assert proc.ran
    assert not proc.threw


def test_event_is_set_on_exception(process_type):
    event = multiprocessing.get_context(process_type).Event()
    proc = DestroyableProcess(process_type, thrower, completion_event=event)
    assert not event.is_set()
    proc.start()
    proc.join()
    assert event.is_set()
    assert not proc.destroyed
    assert not proc.canceled
    assert proc.ran
    assert proc.threw


def test_default_event_is_set_on_completion(process_type):
    proc = DestroyableProcess(process_type, succeeder)
    assert not proc.completion_event.is_set()
    proc.start()
    proc.join()
    assert proc.completion_event.is_set()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""The base class for data model object backends"""

# Standard
from typing import Any, Type, Union
import abc

# Local
from ..base import DataBase

# DataModelBackendBase #########################################################


class DataModelBackendBase(abc.ABC):
    """A base interface class for accessing data from within a given backend
    data layout
    """

    @abc.abstractmethod
    def get_attribute(
        self,
        data_model_class: Type[DataBase],
        name: str,
    ) -> Union[Any, DataBase.OneofFieldVal]:
        """A data model backend must implement this in order to provide the
        frontend view the functionality needed to lazily extract data.

        Args:
            data_model_class (Type[DataBase]): The frontend data model class
                that is accessing this attribute
            name (str): The name of the attribute to access

        Returns:
            value:  Union[Any, OneofFieldVal]
                The extracted attribute value or a OneofFieldVal that wraps the
                field val with an indicator about the oneof field that is set.
        """

    # pylint: disable=unused-argument
    def cache_attribute(self, name: str, value: Any) -> bool:
        """Determine whether or not to cache the given attribute's result on the
        wrapping data model object.

        The base implementation always returns True. Derived classes may opt to
        always return False to fully disable caching, or cache conditionally
        based on the name/value of the individual field.

        Args:
            name (str): The name of the attribute to check
            value (Any): The extracted value

        Returns:
            should_cache:  bool
                True if the value should be cached, False otherwise
        """
        return True



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Common data model enum used for reporting training status
"""

# Standard
from enum import Enum

# Local
from .dataobject import dataobject
from .package import PACKAGE_COMMON


@dataobject(PACKAGE_COMMON)
class TrainingStatus(Enum):
    QUEUED = 1
    RUNNING = 2
    COMPLETED = 3
    CANCELED = 4
    ERRORED = 5

    @property
    def is_terminal(self):
        return self in [
            self.__class__.COMPLETED,
            self.__class__.CANCELED,
            self.__class__.ERRORED,
        ]



"""
A module meant to flex a bit of the protobufs primitive support
"""
# Standard
from dataclasses import field
from typing import Dict, List, Union

# Local
from ...data_model.sample import SampleInputType, SampleOutputType, SampleTask
from caikit.core.data_model.json_dict import JsonDict
from caikit.core.modules import ModuleLoader, ModuleSaver
import caikit.core


@caikit.core.module(
    "00112233-0405-0607-0809-0a0b02dd0e0f", "SampleModule", "0.0.1", SampleTask
)
class SamplePrimitiveModule(caikit.core.ModuleBase):
    def __init__(self, training_params_json_dict=None, training_params_dict={}):
        super().__init__()
        self.training_params_json_dict = training_params_json_dict
        self.training_params_dict = training_params_dict

    @classmethod
    def load(cls, model_path, **kwargs):
        loader = ModuleLoader(model_path)
        config = loader.config
        return cls(
            config["train"]["training_params_json_dict"],
            config["train"]["training_params_dict"],
        )

    def run(
        self,
        sample_input: SampleInputType,
        bool_type: bool = True,
        int_type: int = 42,
        float_type: float = 34.0,
        str_type: str = "moose",
        bytes_type: bytes = b"",
        list_type: List[str] = None,
    ) -> SampleOutputType:
        """This takes in a bunch of primitive types to ensure that we can pass those through the runtime server correctly."""
        assert isinstance(bool_type, bool)
        assert isinstance(int_type, int)
        assert isinstance(float_type, float)
        assert isinstance(str_type, str)
        assert isinstance(bytes_type, bytes)
        return SampleOutputType(
            f"hello: primitives! {self.training_params_json_dict.get('foo').get('bar')} {self.training_params_dict.get('layer_sizes')}"
        )

    def save(self, model_path):
        module_saver = ModuleSaver(
            self,
            model_path=model_path,
        )
        with module_saver:
            config_options = {
                "train": {
                    "training_params_json_dict": self.training_params_json_dict,
                    "training_params_dict": self.training_params_dict,
                },
            }

            module_saver.update_config(config_options)

    @classmethod
    def train(
        cls,
        sample_input: SampleInputType,
        simple_list: List[str],
        union_list: Union[List[str], List[int]],
        union_list2: Union[List[str], List[int], int],
        union_list3: Union[List[str], List[bool]],
        union_list4: Union[List[str], int],
        training_params_json_dict_list: List[JsonDict],
        training_params_json_dict: JsonDict = None,
        training_params_dict: Dict[str, int] = field(default_factory=dict),
        training_params_dict_int: Dict[int, float] = field(default_factory=dict),
    ) -> "SamplePrimitiveModule":
        """Sample training method that produces a trained model"""
        assert type(sample_input) == SampleInputType
        assert isinstance(simple_list, List)
        assert isinstance(union_list, List)
        assert isinstance(union_list2, List)
        assert isinstance(union_list3, List)
        assert isinstance(union_list4, int)
        assert isinstance(training_params_json_dict_list, List)
        assert isinstance(training_params_json_dict, Dict)
        assert isinstance(training_params_dict, Dict)
        assert isinstance(training_params_dict_int, Dict)
        assert training_params_json_dict is not None
        assert len(training_params_json_dict_list) > 0
        assert len(union_list) > 0
        assert len(union_list2) > 0
        assert len(union_list3) > 0
        assert training_params_dict is not None
        assert training_params_dict_int is not None
        return cls(
            training_params_json_dict=training_params_json_dict,
            training_params_dict=training_params_dict,
        )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This script auto-generates the `caikit-runtime.proto` RPC definitions for a
collection of caikit.core derived libraries
"""

# Standard
from typing import Dict, List, Type

# First Party
import alog

# Local
from .rpcs import CaikitRPCBase, ModuleClassTrainRPC, TaskPredictRPC
from caikit.core import ModuleBase, TaskBase
from caikit.core.signature_parsing.module_signature import CaikitMethodSignature

log = alog.use_channel("CREATE-RPCS")

## Globals #####################################################################

TRAIN_FUNCTION_NAME = "train"

## Utilities ###################################################################


def create_inference_rpcs(modules: List[Type[ModuleBase]]) -> List[CaikitRPCBase]:
    """Handles the logic to create all the RPCs for inference"""
    rpcs = []
    task_groups = _group_modules_by_task(modules)

    # Create the RPC for each task
    for task, task_methods in task_groups.items():
        with alog.ContextLog(log.debug, "Generating task RPC for %s", task):
            for streaming_type, method_signatures in task_methods.items():
                input_streaming, output_streaming = streaming_type
                try:
                    rpcs.append(
                        TaskPredictRPC(
                            task, method_signatures, input_streaming, output_streaming
                        )
                    )
                    log.debug("Successfully generated task RPC for %s", task)
                except Exception as err:  # pylint: disable=broad-exception-caught
                    log.warning(
                        "Cannot generate task rpc for %s: %s",
                        task,
                        err,
                        exc_info=True,
                    )

    return rpcs


def create_training_rpcs(modules: List[Type[ModuleBase]]) -> List[CaikitRPCBase]:
    """Handles the logic to create all the RPCs for training"""

    rpcs = []

    for ck_module in modules:
        if not ck_module.TASK_CLASS:
            log.debug("Skipping module %s with no task", ck_module)
            continue

        # If this train function has not been changed from the base, skip it as
        # a module that can't be trained
        #
        # HACK alert! I'm struggling to find the right way to identify this
        #   condition, so for now, we'll use the string repr
        train_fn = getattr(ck_module, TRAIN_FUNCTION_NAME)
        if str(train_fn).startswith(f"<bound method ModuleBase.{TRAIN_FUNCTION_NAME}"):
            log.debug(
                "Skipping train API for %s with no %s function",
                ck_module,
                TRAIN_FUNCTION_NAME,
            )
            continue

        signature = ck_module.TRAIN_SIGNATURE
        log.debug(
            "Function signature for %s::%s [%s -> %s]",
            ck_module,
            TRAIN_FUNCTION_NAME,
            signature.parameters,
            signature.return_type,
        )
        with alog.ContextLog(log.debug, "Generating train RPC for %s", ck_module):
            try:
                rpcs.append(ModuleClassTrainRPC(signature))
                log.debug("Successfully generated train RPC for %s", ck_module)
            except Exception as err:  # pylint: disable=broad-exception-caught
                log.warning(
                    "Cannot generate train rpc for %s: %s",
                    ck_module,
                    err,
                    exc_info=True,
                )
    return rpcs


def _group_modules_by_task(
    modules: List[Type[ModuleBase]],
) -> Dict[Type[TaskBase], List[CaikitMethodSignature]]:
    task_groups = {}
    for ck_module in modules:
        if ck_module.TASK_CLASS:
            ck_module_task_name = ck_module.TASK_CLASS.__name__
            if ck_module_task_name is not None:
                for (
                    input_streaming,
                    output_streaming,
                    signature,
                ) in ck_module._INFERENCE_SIGNATURES:
                    task_groups.setdefault(ck_module.TASK_CLASS, {}).setdefault(
                        (input_streaming, output_streaming), []
                    ).append(signature)
    return task_groups



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General model train servicer tests - run for both subprocess and local"""
# Standard
from contextlib import contextmanager
import json
import os
import tempfile
import uuid

# Third Party
import grpc
import pytest

# Local
from caikit.core import MODEL_MANAGER
from caikit.core.data_model import TrainingStatus
from caikit.runtime.protobufs import process_pb2
from caikit.runtime.servicers.model_train_servicer import ModelTrainServicerImpl
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from tests.conftest import temp_config
from tests.fixtures import Fixtures
import sample_lib


@pytest.fixture
def sample_model_train_servicer(sample_train_service) -> ModelTrainServicerImpl:
    servicer = ModelTrainServicerImpl(training_service=sample_train_service)
    yield servicer


@pytest.fixture
def output_dir() -> str:
    with tempfile.TemporaryDirectory() as workdir:
        yield workdir


# temporarily clear all messages from servicer
@contextmanager
def clear_messages_from_servicer(servicer):
    messages = servicer._training_service.messages
    servicer._training_service.messages = []
    try:
        yield
    finally:
        servicer._training_service.messages = messages


@pytest.fixture(autouse=True, params=[True, False])
def set_train_location(request):
    """This fixture ensures that all tests in this file will be run with both
    subprocess and local training styles
    """
    with temp_config({"training": {"use_subprocess": request.param}}):
        yield


#####################################################################
# Error tests


def test_model_train_no_training_params_raises(sample_model_train_servicer, output_dir):
    training_id = str(uuid.uuid4())
    model_train_request = process_pb2.ProcessRequest(
        trainingID=training_id,
        customTrainingID=str(uuid.uuid4()),
        request_dict={
            "train_module": "00110203-0405-0607-0809-0a0b02dd0e0f",
        },
        training_input_dir="training_input_dir",
        training_output_dir=os.path.join(output_dir, training_id),
    )
    with pytest.raises(CaikitRuntimeException):
        context = Fixtures.build_context("foo")
        sample_model_train_servicer.Run(model_train_request, context)


def test_model_train_no_train_module_raises(sample_model_train_servicer, output_dir):
    training_id = str(uuid.uuid4())
    model_train_request = process_pb2.ProcessRequest(
        trainingID=training_id,
        customTrainingID=str(uuid.uuid4()),
        request_dict={
            "training_params": '{"model_name": "abc", "parameters": {"training_data": [1]}}',
        },
        training_input_dir="training_input_dir",
        training_output_dir=os.path.join(output_dir, training_id),
    )
    with pytest.raises(CaikitRuntimeException):
        context = Fixtures.build_context("foo")
        sample_model_train_servicer.Run(model_train_request, context)


def test_model_train_incorrect_train_params_raises(
    sample_model_train_servicer, output_dir
):
    training_id = str(uuid.uuid4())
    model_train_request = process_pb2.ProcessRequest(
        trainingID=training_id,
        customTrainingID=str(uuid.uuid4()),
        request_dict={
            "train_module": "00110203-0405-0607-0809-0a0b02dd0e0f",
            "training_params": '{"model_name": "abc", "parameters": {"training_data": "blah"}}',
        },
        training_input_dir="training_input_dir",
        training_output_dir=os.path.join(output_dir, training_id),
    )
    with pytest.raises(CaikitRuntimeException):
        context = Fixtures.build_context("foo")
        sample_model_train_servicer.Run(model_train_request, context)


def test_model_train_incorrect_train_module_raises(
    sample_model_train_servicer, output_dir
):
    training_id = str(uuid.uuid4())
    model_train_request = process_pb2.ProcessRequest(
        trainingID=training_id,
        customTrainingID=str(uuid.uuid4()),
        request_dict={
            "train_module": "random_id",
        },
        training_input_dir="training_input_dir",
        training_output_dir=os.path.join(output_dir, training_id),
    )
    with pytest.raises(CaikitRuntimeException) as e:
        context = Fixtures.build_context("foo")
        sample_model_train_servicer.Run(model_train_request, context)

    assert "Model Train not able to parse module for this Train Request" in str(e.value)
    assert e.value.status_code == grpc.StatusCode.INVALID_ARGUMENT


def test_model_train_incorrect_train_request_raises(
    sample_model_train_servicer, output_dir
):
    training_id = str(uuid.uuid4())
    model_train_request = process_pb2.ProcessRequest(
        trainingID=training_id,
        customTrainingID=str(uuid.uuid4()),
        request_dict={
            "train_module": "00110203-0405-0607-0809-0a0b02dd0e0f",
            "training_params": '{"model_name": "abc", "parameters": {"training_data": [1]}}',
        },
        training_input_dir="training_input_dir",
        training_output_dir=os.path.join(output_dir, training_id),
    )
    with clear_messages_from_servicer(sample_model_train_servicer):
        with pytest.raises(CaikitRuntimeException) as e:
            context = Fixtures.build_context("foo")
            sample_model_train_servicer.Run(model_train_request, context)

        assert "Model Train not able to create a request" in str(e.value)
        assert e.value.status_code == grpc.StatusCode.INTERNAL


def test_model_train_validation_error_raises(sample_model_train_servicer, output_dir):
    """Test that if there is validation error we are able to raise it"""

    training_id = str(uuid.uuid4())
    model_train_request = process_pb2.ProcessRequest(
        trainingID=training_id,
        customTrainingID=str(uuid.uuid4()),
        request_dict={
            "train_module": "00110203-0405-0607-0809-0a0b02dd0e0f",
            "training_params": '{"model_name": "abc", "parameters": {"training_data": [1], "batch_size": 999}}',
        },
        training_input_dir="training_input_dir",
        training_output_dir=os.path.join(output_dir, training_id),
    )
    with pytest.raises(CaikitRuntimeException):
        context = Fixtures.build_context("foo")
        sample_model_train_servicer.Run(model_train_request, context)


#####################################################################
# Normal tests
@pytest.mark.parametrize("output_in_config", (True, False))
@pytest.mark.parametrize("save_with_id", (True, False))
@pytest.mark.parametrize("use_custom_id", (True, False))
def test_model_train_sample_widget(
    sample_model_train_servicer,
    output_dir,
    output_in_config,
    save_with_id,
    use_custom_id,
):
    """This test tests end-to-end training. It includes verifying that the model
    is saved in the right place. The place where the model is saved comes from:

    1. get_config().runtime.training.output_dir
    2. request.training_output_dir

    IN THAT ORDER!
    """
    training_id = str(uuid.uuid4())
    custom_id = f"custom-{uuid.uuid4()}" if use_custom_id else None
    model_name = "abc"

    # If getting the output path from config, use a bogus dir in the request to
    # ensure that the config is preferred
    if output_in_config:
        config_output_dir = output_dir
        req_output_dir = "/some/bad/path"
    # Otherwise the config must be set to an empty string so the request is used
    else:
        config_output_dir = ""
        req_output_dir = output_dir

    with temp_config(
        {
            "runtime": {
                "training": {
                    "output_dir": config_output_dir,
                    "save_with_id": save_with_id,
                },
            },
        },
        "merge",
    ):
        model_train_request = process_pb2.ProcessRequest(
            trainingID=training_id,
            customTrainingID=custom_id,
            request_dict={
                "train_module": "00110203-0405-0607-0809-0a0b02dd0e0f",
                "training_params": json.dumps(
                    {
                        "model_name": model_name,
                        "parameters": {
                            "training_data": {
                                "jsondata": {
                                    "data": [
                                        sample_lib.data_model.SampleTrainingType(
                                            number=1
                                        ).to_dict()
                                    ]
                                },
                            },
                        },
                    }
                ),
            },
            training_input_dir="training_input_dir",
            training_output_dir=req_output_dir,
        )
        context = Fixtures.build_context("test-any-unresponsive-model")
        training_response = sample_model_train_servicer.Run(
            model_train_request, context
        )
        if save_with_id:
            if use_custom_id:
                model_save_path = os.path.join(output_dir, custom_id, model_name)
            else:
                model_save_path = os.path.join(output_dir, training_id, model_name)
        else:
            model_save_path = os.path.join(output_dir, model_name)
        assert os.path.isdir(model_save_path)

    # Make sure that the request completed synchronously
    model_future = MODEL_MANAGER.get_model_future(custom_id or training_id)
    assert model_future.get_info().status == TrainingStatus.COMPLETED

    # Make sure that the return object looks right
    assert isinstance(training_response, process_pb2.ProcessResponse)
    assert training_response == process_pb2.ProcessResponse(
        trainingID=training_id,
        customTrainingID=model_train_request.customTrainingID,
    )


def test_files_from_training_input_dir_are_used(
    sample_model_train_servicer, output_dir
):
    """The MT request comes with a `training_input_dir` field.
    If datastreams reference files, they may exist inside this dynamically-created directory.
    The servicer should detect this and direct datastreams to that directory.
    """
    training_id = str(uuid.uuid4())

    training_output_dir = os.path.join(output_dir, training_id)
    training_input_dir = os.path.join(output_dir, training_id, "inputs")
    input_file_name = "data.json"

    os.makedirs(training_input_dir, exist_ok=True)
    with open(os.path.join(training_input_dir, input_file_name), "w") as f:
        json.dump([sample_lib.data_model.SampleTrainingType(number=1).to_dict()], f)

    model_train_request = process_pb2.ProcessRequest(
        trainingID=training_id,
        request_dict={
            "train_module": "00110203-0405-0607-0809-0a0b02dd0e0f",
            "training_params": json.dumps(
                {
                    "model_name": "abc",
                    "parameters": {
                        "training_data": {
                            "file": {
                                "filename": input_file_name  # This is relative to training_input_dir
                            },
                        },
                    },
                }
            ),
        },
        training_input_dir=training_input_dir,
        training_output_dir=training_output_dir,
    )
    context = Fixtures.build_context("test-any-unresponsive-model")
    training_response = sample_model_train_servicer.Run(model_train_request, context)
    assert os.path.isdir(training_output_dir)

    # Make sure that the training succeeded
    model_future = MODEL_MANAGER.get_model_future(training_response.trainingID)
    assert model_future.get_info().status == TrainingStatus.COMPLETED



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the rpc objects that hold our in-memory representation of
what an RPC for a service looks like"""
# Standard
from typing import Iterable, Union
import uuid

# Local
from caikit.core import ModuleBase, TaskBase
from caikit.core.data_model import DataStream
from caikit.runtime.service_generation.rpcs import ModuleClassTrainRPC, TaskPredictRPC
from sample_lib.data_model import SampleInputType, SampleOutputType
import caikit.core


def test_task_inference_multiples_modules_rpc():
    @caikit.core.task(
        unary_parameters={"sample_input": SampleInputType},
        streaming_parameters={"sample_inputs": Iterable[SampleInputType]},
        unary_output_type=SampleOutputType,
        streaming_output_type=Iterable[SampleOutputType],
    )
    class MultiModTask(TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()), name="testmod1", version="9.9.9", task=MultiModTask
    )
    class TestModule1(ModuleBase):
        @MultiModTask.taskmethod(input_streaming=True)
        def run_stream_in(
            self, sample_inputs: DataStream[SampleInputType]
        ) -> SampleOutputType:
            pass

    @caikit.core.module(
        id=str(uuid.uuid4()), name="testmod2", version="9.9.9", task=MultiModTask
    )
    class TestModule2(ModuleBase):
        @MultiModTask.taskmethod(input_streaming=True)
        def run_stream_in(
            self, sample_inputs: DataStream[SampleInputType]
        ) -> SampleOutputType:
            pass

    rpc = TaskPredictRPC(
        task=MultiModTask,
        method_signatures=[
            TestModule1.get_inference_signature(
                input_streaming=True, output_streaming=False
            ),
            TestModule2.get_inference_signature(
                input_streaming=True, output_streaming=False
            ),
        ],
        input_streaming=True,
        output_streaming=False,
    )
    assert rpc.request.name == "ClientStreamingMultiModTaskRequest"
    assert rpc.request.triples == [(SampleInputType, "sample_inputs", 1)]

    data_model = rpc.create_request_data_model(package_name="blah")
    assert data_model is not None

    assert rpc.name == "ClientStreamingMultiModTaskPredict"


def test_task_inference_rpc_with_client_streaming():
    @caikit.core.task(
        unary_parameters={"sample_input": SampleInputType},
        streaming_parameters={"sample_inputs": Iterable[SampleInputType]},
        unary_output_type=SampleOutputType,
        streaming_output_type=Iterable[SampleOutputType],
    )
    class TestTask1(TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()), name="testest", version="9.9.9", task=TestTask1
    )
    class TestModule(ModuleBase):
        @TestTask1.taskmethod(input_streaming=True)
        def run_stream_in(
            self, sample_inputs: DataStream[SampleInputType]
        ) -> SampleOutputType:
            pass

    rpc = TaskPredictRPC(
        task=TestTask1,
        method_signatures=[
            TestModule.get_inference_signature(
                input_streaming=True, output_streaming=False
            )
        ],
        input_streaming=True,
        output_streaming=False,
    )
    assert rpc.request.name == "ClientStreamingTestTask1Request"
    assert rpc.request.triples == [(SampleInputType, "sample_inputs", 1)]

    data_model = rpc.create_request_data_model(package_name="blah")
    assert data_model is not None

    assert rpc.name == "ClientStreamingTestTask1Predict"


def test_task_inference_rpc_with_streaming():
    @caikit.core.task(
        unary_parameters={"sample_input": SampleInputType},
        streaming_parameters={"sample_inputs": Iterable[SampleInputType]},
        unary_output_type=SampleOutputType,
        streaming_output_type=Iterable[SampleOutputType],
    )
    class TestTask(TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()), name="testest", version="9.9.9", task=TestTask
    )
    class TestModule(ModuleBase):
        @TestTask.taskmethod()
        def run(self, sample_input: SampleInputType) -> SampleOutputType:
            pass

        @TestTask.taskmethod(output_streaming=True)
        def run_stream_out(
            self, sample_input: SampleInputType
        ) -> DataStream[SampleOutputType]:
            pass

        @TestTask.taskmethod(input_streaming=True)
        def run_stream_in(
            self, sample_inputs: DataStream[SampleInputType]
        ) -> SampleOutputType:
            pass

        @TestTask.taskmethod(input_streaming=True, output_streaming=True)
        def run_stream_bidi(
            self, sample_inputs: Iterable[SampleInputType]
        ) -> DataStream[SampleOutputType]:
            pass

    # Unary
    rpc = TaskPredictRPC(
        task=TestTask,
        method_signatures=[
            TestModule.get_inference_signature(
                input_streaming=False, output_streaming=False
            )
        ],
    )

    assert rpc.name == "TestTaskPredict"

    # # Server streaming
    rpc = TaskPredictRPC(
        task=TestTask,
        method_signatures=[
            TestModule.get_inference_signature(
                input_streaming=False, output_streaming=True
            )
        ],
        input_streaming=False,
        output_streaming=True,
    )

    assert rpc.name == "ServerStreamingTestTaskPredict"

    # Client streaming
    rpc = TaskPredictRPC(
        task=TestTask,
        method_signatures=[
            TestModule.get_inference_signature(
                input_streaming=True, output_streaming=False
            )
        ],
        input_streaming=True,
        output_streaming=False,
    )

    assert rpc.name == "ClientStreamingTestTaskPredict"

    # Bidi streaming
    rpc = TaskPredictRPC(
        task=TestTask,
        method_signatures=[
            TestModule.get_inference_signature(
                input_streaming=True, output_streaming=True
            )
        ],
        input_streaming=True,
        output_streaming=True,
    )

    assert rpc.name == "BidiStreamingTestTaskPredict"


def test_task_inference_rpc_with_all_optional_params():
    @caikit.core.task(
        required_parameters={"str_val": str}, output_type=SampleOutputType
    )
    class TestTask(TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()), name="testest", version="9.9.9", task=TestTask
    )
    class TestModule(ModuleBase):
        def run(self, str_val="I have a default") -> SampleOutputType:
            pass

    rpc = TaskPredictRPC(
        task=TestTask,
        method_signatures=[TestModule.RUN_SIGNATURE],
    )

    data_model = rpc.create_request_data_model(package_name="blah")
    assert data_model is not None

    assert rpc.name == "TestTaskPredict"


def test_module_train_rpc():
    @caikit.core.task(
        required_parameters={"str_val": str}, output_type=SampleOutputType
    )
    class TestTask(TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()), name="testest", version="9.9.9", task=TestTask
    )
    class TestModule(ModuleBase):
        def run(self, str_val: str) -> SampleOutputType:
            pass

        @classmethod
        def train(cls, int_val: int, str_val: str) -> "TestModule":
            pass

    rpc = ModuleClassTrainRPC(method_signature=TestModule.TRAIN_SIGNATURE)

    data_model = rpc.create_request_data_model(package_name="blah")
    assert data_model is not None

    assert rpc.name == "TestTaskTestModuleTrain"

    # Training RPCs nest the actual training params from the `.train` signature
    assert hasattr(data_model, "model_name")
    assert hasattr(data_model, "output_path")
    assert hasattr(data_model, "parameters")

    training_message = data_model.from_json(
        {"model_name": "any_model_name", "parameters": {"int_val": 1, "str_val": "foo"}}
    )

    assert training_message.parameters.int_val == 1
    assert training_message.parameters.str_val == "foo"



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The MultiModelFinder configures a set of other model finders that will be used
in sequence to try loading models.

Configuration for MultiModelFinder lives under the config as follows:

model_management:
    finders:
        <finder name>:
            type: MULTI
            config:
                # Sequence of other finder names to use in priority order
                finder_sequence:
                    - other_finder1
                    - other_finder2
"""
# Standard
from typing import Optional

# First Party
import aconfig
import alog

# Local
from ...config import get_config
from ..exceptions import error_handler
from ..modules import ModuleConfig
from .model_finder_base import ModelFinderBase

# NOTE: Top-level import done so that global MODEL_MANAGER can be used at
#   construction time without incurring a circular dependency
import caikit.core

log = alog.use_channel("MFIND")
error = error_handler.get(log)


class MultiModelFinder(ModelFinderBase):
    __doc__ = __doc__

    name = "MULTI"

    def __init__(self, config: aconfig.Config, instance_name: str):
        """Initialize with the sequence of finders to use"""
        self._instance_name = instance_name
        finder_priority = config.finder_priority
        error.type_check(
            "<COR47518221E>",
            list,
            finder_priority=finder_priority,
        )
        error.type_check_all(
            "<COR47518222E>",
            str,
            finder_priority=finder_priority,
        )
        error.value_check(
            "<COR05042343E>",
            finder_priority,
            "Must provide at least one valid finder",
        )
        config_finders = get_config().model_management.finders
        invalid_finders = [
            finder for finder in finder_priority if finder not in config_finders
        ]
        error.value_check(
            "<COR68252034E>",
            not invalid_finders,
            "Invalid finders given in finder_priority: {}",
            invalid_finders,
        )
        error.value_check(
            "<COR54613971E>",
            self._instance_name not in finder_priority,
            "Cannot include self in multi finder priority",
        )
        model_manager = config.model_manager or caikit.core.MODEL_MANAGER
        log.debug2("Setting up %s with finder priority: %s", self.name, finder_priority)
        self._finders = [model_manager.get_finder(finder) for finder in finder_priority]

    def find_model(
        self,
        model_path: str,
        **kwargs,
    ) -> Optional[ModuleConfig]:
        """Iterate through the sequence of finders and return the first one that
        succeeds
        """
        for idx, finder in enumerate(self._finders):
            log.debug2(
                "Trying to find %s with finder %d of type %s",
                model_path,
                idx,
                finder.name,
            )
            try:
                module_config = finder.find_model(model_path, **kwargs)
                if module_config:
                    log.debug(
                        "Found %s with finder %d of type %s",
                        model_path,
                        idx,
                        finder.name,
                    )
                    return module_config
                log.debug2(
                    "Finder %d of type %s unable to find %s",
                    idx,
                    finder.name,
                    model_path,
                )
            except Exception as err:  # pylint: disable=broad-exception-caught
                log.debug2(
                    "Finder %d of type %s failed to load %s: %s",
                    idx,
                    finder.name,
                    model_path,
                    err,
                )
                log.debug4("Finder error", exc_info=True)

        # No finder succeeded
        log.warning("Unable to find %s with any finder", model_path)
        return None



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the service factory"""
# Standard
from pathlib import Path
import tempfile

# Third Party
from google.protobuf.message import Message
import pytest

# Local
from caikit.core.data_model import render_dataobject_protos
from caikit.runtime.service_factory import (
    ServicePackage,
    ServicePackageFactory,
    get_inference_request,
    get_train_params,
    get_train_request,
)
from sample_lib import SampleModule
from sample_lib.data_model import SampleInputType, SampleOutputType
from sample_lib.modules import ListModule, OtherModule
from tests.conftest import temp_config
from tests.core.helpers import MockBackend
from tests.data_model_helpers import reset_global_protobuf_registry, temp_dpool
from tests.runtime.conftest import sample_inference_service, sample_train_service
import caikit

## Helpers #####################################################################


@pytest.fixture
def clean_data_model(sample_inference_service, sample_train_service):
    """Set up a temporary descriptor pool that inherits all explicitly created
    dataobjects from the global, but skips dynamically created ones from
    datastreamsource.
    """
    with reset_global_protobuf_registry():
        with temp_dpool(
            inherit_global=True,
            skip_inherit=[".*sampletask.*\.proto"],
        ) as dpool:
            yield dpool


def validate_package_with_override(
    service_type: ServicePackageFactory.ServiceType,
    domain_override: str,
    package_override: str,
) -> ServicePackage:
    """Construct and validate train or infer service package with domain and package path override

    The config should be set up for domain and/or package override, before this function is called.

    Args:
        service_type (ServicePackageFactory.ServiceType) : The type of service to build,
            only TRAINING and INFERENCE are supported
        domain_override (str): The domain to validate (maybe overridden value, or the default)
        package_override (str): The package path to validate (maybe overridden value, or the default)

    Returns: The service package that was created, for potential subsequent reuse
    """
    assert service_type in {
        ServicePackageFactory.ServiceType.TRAINING,
        ServicePackageFactory.ServiceType.INFERENCE,
    }
    svc = ServicePackageFactory.get_service_package(service_type)

    service_name = (
        f"{domain_override}TrainingService"
        if service_type == ServicePackageFactory.ServiceType.TRAINING
        else f"{domain_override}Service"
    )
    assert svc.service.__name__ == service_name
    assert svc.descriptor.full_name == f"{package_override}.{service_name}"
    for message_name in [
        x for x in svc.messages.__dict__.keys() if not x.startswith("_")
    ]:
        message: Message = getattr(svc.messages, message_name)
        assert message.DESCRIPTOR.full_name == f"{package_override}.{message_name}"

    return svc


### Private method tests #############################################################


MODULE_LIST = [
    module_class
    for module_class in caikit.core.registries.module_registry().values()
    if module_class.__module__.partition(".")[0] == "sample_lib"
]


### Test ServicePackageFactory._get_and_filter_modules function
def test_get_and_filter_modules_respects_excluded_task_type():
    with temp_config(
        {
            "runtime": {
                "service_generation": {"task_types": {"excluded": ["SampleTask"]}}
            }
        }
    ) as cfg:
        clean_modules = ServicePackageFactory._get_and_filter_modules(cfg, "sample_lib")
        assert "SampleModule" not in str(clean_modules)


def test_get_and_filter_modules_respects_excluded_modules():
    assert "InnerModule" in str(MODULE_LIST)
    with temp_config(
        {
            "runtime": {
                "service_generation": {
                    "module_guids": {"excluded": [ListModule.MODULE_ID]}
                }
            }
        }
    ) as cfg:
        clean_modules = ServicePackageFactory._get_and_filter_modules(cfg, "sample_lib")
        assert "ListModule" not in str(clean_modules)
        assert "SampleModule" in str(clean_modules)
        assert "OtherModule" in str(clean_modules)


def test_get_and_filter_modules_respects_excluded_modules_and_excluded_task_type():
    assert "InnerModule" in str(MODULE_LIST)
    with temp_config(
        {
            "runtime": {
                "service_generation": {
                    "module_guids": {"excluded": [ListModule.MODULE_ID]},
                    "task_types": {"excluded": ["OtherTask"]},
                }
            }
        }
    ) as cfg:
        clean_modules = ServicePackageFactory._get_and_filter_modules(cfg, "sample_lib")
        assert "ListModule" not in str(clean_modules)
        assert "OtherModule" not in str(clean_modules)
        assert "SampleModule" in str(clean_modules)


def test_get_and_filter_modules_respects_included_modules_and_included_task_types():
    with temp_config(
        {
            "runtime": {
                "service_generation": {
                    "module_guids": {"included": [ListModule.MODULE_ID]},
                    "task_types": {"included": ["OtherTask"]},
                }
            }
        }
    ) as cfg:
        clean_modules = ServicePackageFactory._get_and_filter_modules(cfg, "sample_lib")
        assert len(clean_modules) == 2
        assert "OtherModule" in str(clean_modules)
        assert "ListModule" in str(clean_modules)


def test_get_and_filter_modules_respects_included_modules():
    with temp_config(
        {
            "runtime": {
                "service_generation": {
                    "module_guids": {"included": [ListModule.MODULE_ID]},
                }
            }
        }
    ) as cfg:
        clean_modules = ServicePackageFactory._get_and_filter_modules(cfg, "sample_lib")
        assert len(clean_modules) == 1
        assert "ListModule" in str(clean_modules)
        assert "SampleModule" not in str(clean_modules)


def test_get_and_filter_modules_respects_included_task_types():
    with temp_config(
        {
            "runtime": {
                "service_generation": {
                    "task_types": {"included": ["SampleTask"]},
                }
            }
        }
    ) as cfg:
        clean_modules = ServicePackageFactory._get_and_filter_modules(cfg, "sample_lib")
        assert "SampleModule" in str(clean_modules)
        assert "OtherModule" not in str(clean_modules)
        # InnerModule has no task
        assert "InnerModule" not in str(clean_modules)


def test_get_and_filter_modules_respects_included_task_types_and_excluded_modules():
    with temp_config(
        {
            "runtime": {
                "service_generation": {
                    "task_types": {"included": ["SampleTask"]},
                    "module_guids": {"excluded": [ListModule.MODULE_ID]},
                }
            }
        }
    ) as cfg:
        clean_modules = ServicePackageFactory._get_and_filter_modules(cfg, "sample_lib")
        assert "SampleModule" in str(clean_modules)
        assert "ListModule" not in str(clean_modules)


def test_override_domain(clean_data_model):
    """
    Test override of gRPC domain generation from config.
    The feature allows achieving backwards compatibility with earlier gRPC client.
    """
    domain_override = "OverrideDomainName"
    with temp_config(
        {
            "runtime": {
                "library": "sample_lib",
                "service_generation": {
                    "task_types": {"included": ["SampleTask"]},
                    "domain": domain_override,
                },
            }
        }
    ) as cfg:
        # Changing the domain still effects the default package name.
        # But if you override the package, it overrides the package name completely (see next test).
        validate_package_with_override(
            ServicePackageFactory.ServiceType.INFERENCE,
            domain_override,
            f"caikit.runtime.{domain_override}",
        )
        validate_package_with_override(
            ServicePackageFactory.ServiceType.TRAINING,
            domain_override,
            f"caikit.runtime.{domain_override}",
        )

        # Just double-check that basics are good.
        clean_modules = ServicePackageFactory._get_and_filter_modules(cfg, "sample_lib")
        assert "SampleModule" in str(clean_modules)


def test_override_package(clean_data_model):
    """
    Test override of gRPC package generation from config.
    The feature allows achieving backwards compatibility with earlier gRPC client.
    """
    package_override = "foo.runtime.yada.v0"
    with temp_config(
        {
            "runtime": {
                "library": "sample_lib",
                "service_generation": {
                    "task_types": {"included": ["SampleTask"]},
                    "package": package_override,
                },
            }
        }
    ) as cfg:
        validate_package_with_override(
            ServicePackageFactory.ServiceType.INFERENCE, "SampleLib", package_override
        )
        validate_package_with_override(
            ServicePackageFactory.ServiceType.TRAINING, "SampleLib", package_override
        )

        # Just double-check that basics are good.
        clean_modules = ServicePackageFactory._get_and_filter_modules(cfg, "sample_lib")
        assert "SampleModule" in str(clean_modules)


def test_override_package_and_domain_with_proto_gen(clean_data_model):
    """
    Test override of both package and domain, to make sure they work together, and
    additionally test the proto generation.
    The feature allows achieving backwards compatibility with earlier gRPC client.
    """
    package_override = "foo.runtime.yada.v0"
    domain_override = "OverrideDomainName"
    with temp_config(
        {
            "runtime": {
                "library": "sample_lib",
                "service_generation": {
                    "task_types": {"included": ["SampleTask"]},
                    "package": package_override,
                    "domain": domain_override,
                },
            }
        }
    ) as cfg:
        inf_svc = validate_package_with_override(
            ServicePackageFactory.ServiceType.INFERENCE,
            domain_override,
            package_override,
        )
        train_svc = validate_package_with_override(
            ServicePackageFactory.ServiceType.TRAINING,
            domain_override,
            package_override,
        )

        # Just double-check that basics are good.
        clean_modules = ServicePackageFactory._get_and_filter_modules(cfg, "sample_lib")
        assert "SampleModule" in str(clean_modules)

        # Full check on proto generation
        with tempfile.TemporaryDirectory() as output_dir:
            render_dataobject_protos(output_dir)
            inf_svc.service.write_proto_file(output_dir)
            train_svc.service.write_proto_file(output_dir)

            output_dir_path = Path(output_dir)
            for proto_file in output_dir_path.glob("*.proto"):
                with open(proto_file, "rb") as f:
                    lines = f.readlines()
                    for line in lines:
                        line = line.strip().decode("utf-8")
                        if line.startswith("package "):
                            package_name = line.split()[1]
                            assert package_name[-1] == ";"
                            package_name = package_name[0:-1]
                            root_package_name = package_name.split(".")[0]
                            if root_package_name not in {"caikit_data_model", "caikit"}:
                                assert package_name == package_override
                        elif line.startswith("service "):
                            service_name = line.split()[1]
                            domain_override_lower = domain_override.lower()
                            if proto_file.name.startswith(domain_override_lower):
                                if proto_file.stem.endswith("trainingservice"):
                                    assert (
                                        service_name
                                        == f"{domain_override}TrainingService"
                                    )
                                else:
                                    assert service_name == f"{domain_override}Service"


def test_backend_modules_included_in_service_generation(
    clean_data_model, reset_globals
):
    # Add a new backend module for the good ol' `SampleModule`
    @caikit.module(backend_type=MockBackend.backend_type, base_module=SampleModule)
    class NewBackendModule(caikit.core.ModuleBase):
        def run(
            self, sample_input: SampleInputType, backend_param: str
        ) -> SampleOutputType:
            pass

    inference_service = ServicePackageFactory.get_service_package(
        ServicePackageFactory.ServiceType.INFERENCE
    )
    predict_class = get_inference_request(SampleModule.TASK_CLASS)
    sample_task_request = predict_class().to_proto()

    # Check that the new parameter defined in this backend module exists in the service
    assert "backend_param" in sample_task_request.DESCRIPTOR.fields_by_name.keys()


def test_get_inference_request_throws_wrong_type(runtime_grpc_server):
    with pytest.raises(TypeError) as e:
        get_inference_request(task_or_module_class="something random")
    assert "subclass check failed" in e.value.args[0]


def test_get_inference_request(runtime_grpc_server):
    """Test that we are able to get inference request DM with either module or task class"""
    assert get_inference_request(SampleModule).__name__ == "SampleTaskRequest"
    assert (
        get_inference_request(SampleModule.TASK_CLASS).__name__ == "SampleTaskRequest"
    )
    assert (
        get_inference_request(SampleModule, output_streaming=True).__name__
        == "ServerStreamingSampleTaskRequest"
    )
    assert (
        get_inference_request(SampleModule.TASK_CLASS, output_streaming=True).__name__
        == "ServerStreamingSampleTaskRequest"
    )
    assert (
        get_inference_request(
            SampleModule, input_streaming=True, output_streaming=True
        ).__name__
        == "BidiStreamingSampleTaskRequest"
    )
    assert (
        get_inference_request(
            SampleModule.TASK_CLASS, input_streaming=True, output_streaming=True
        ).__name__
        == "BidiStreamingSampleTaskRequest"
    )


def test_get_train_request_throws_wrong_type(runtime_grpc_server):
    with pytest.raises(TypeError) as e:
        get_train_request("not_a_module")
    assert "subclass check failed" in e.value.args[0]


def test_get_train_request(runtime_grpc_server):
    assert (
        get_train_request(SampleModule).__name__ == "SampleTaskSampleModuleTrainRequest"
    )
    assert get_train_request(OtherModule).__name__ == "OtherTaskOtherModuleTrainRequest"


def test_get_train_params_throws_wrong_type(runtime_grpc_server):
    with pytest.raises(TypeError) as e:
        get_train_params("not_a_module")
    assert "subclass check failed" in e.value.args[0]


def test_get_train_params(runtime_grpc_server):
    assert (
        get_train_params(SampleModule).__name__
        == "SampleTaskSampleModuleTrainParameters"
    )
    assert (
        get_train_params(OtherModule).__name__ == "OtherTaskOtherModuleTrainParameters"
    )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for image data model and its backend/views."""
# Standard
from pathlib import Path
import io
import os

# Third Party
from PIL import Image as PILImage
import numpy as np
import pytest

# Local
from caikit.interfaces import vision as v

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), "..", "..", "fixtures")

HAPPY_IMAGE_PATH = os.path.join(FIXTURES_DIR, "data", "sample_image.png")
SAD_IMAGE_PATH = "does/not/exist"

with open(HAPPY_IMAGE_PATH, "rb") as raw_img:
    RAW_IMG_BYTES = raw_img.read()
    PIL_IMG_DATA = PILImage.open(io.BytesIO(RAW_IMG_BYTES))
NP_UINT8_DATA = np.asarray(PIL_IMG_DATA)

IMAGE_DATA_KEY = "image_data"
VALID_INITIALIZATIONS = [
    # A path to an image
    {IMAGE_DATA_KEY: HAPPY_IMAGE_PATH},
    # Pathlib object to an image
    {IMAGE_DATA_KEY: Path(HAPPY_IMAGE_PATH)},
    # A numpy array of type np.uint8 [0-255]
    {IMAGE_DATA_KEY: NP_UINT8_DATA},
    # A PIL image
    {IMAGE_DATA_KEY: PIL_IMG_DATA},
    # Raw bytes representing a compressed image
    {IMAGE_DATA_KEY: RAW_IMG_BYTES},
]

INVALID_INITIALIZATIONS = [
    # Numpy img cast to float32
    {IMAGE_DATA_KEY: NP_UINT8_DATA.astype(np.float32)},
]

### Initialization tests
@pytest.mark.parametrize("happy_kwargs", VALID_INITIALIZATIONS)
def test_valid_image_initializations_with_kwargs(happy_kwargs):
    """Test that we can build an Caikit Image data model object directly [kwargs]."""
    im = v.data_model.Image(**happy_kwargs)
    assert isinstance(im, v.data_model.Image)
    # Ensure that no matter what, we always have the same data
    numpy_im = im.as_numpy()
    assert isinstance(numpy_im, np.ndarray)
    assert np.allclose(numpy_im, NP_UINT8_DATA)


@pytest.mark.parametrize("happy_kwargs", VALID_INITIALIZATIONS)
def test_valid_image_initializations_with_positional_args(happy_kwargs):
    """Test that we can build an Caikit Image data model object directly [positional args]."""
    im = v.data_model.Image(happy_kwargs[IMAGE_DATA_KEY])
    assert isinstance(im, v.data_model.Image)
    # Ensure that no matter what, we always have the same data
    numpy_im = im.as_numpy()
    assert isinstance(numpy_im, np.ndarray)
    assert np.allclose(numpy_im, NP_UINT8_DATA)


@pytest.mark.parametrize("sad_kwargs", INVALID_INITIALIZATIONS)
def test_invalid_initializations_with_kwargs(sad_kwargs):
    """Test that we get a ValueError if we try to init with unsupported types [kwargs]."""
    with pytest.raises(ValueError):
        v.data_model.Image(**sad_kwargs)


@pytest.mark.parametrize("sad_kwargs", INVALID_INITIALIZATIONS)
def test_invalid_initializations_with_positional_args(sad_kwargs):
    """Test that we get a ValueError if we try to init with unsupported types [positional args]."""
    with pytest.raises(ValueError):
        v.data_model.Image(sad_kwargs[IMAGE_DATA_KEY])


def test_bad_path_initialization():
    """Test that we get a FileNotFoundError if we try to load a bad path."""
    with pytest.raises(FileNotFoundError):
        v.data_model.Image(image_data=SAD_IMAGE_PATH)


def test_bad_pathlib_initialization():
    """Test that we get a FileNotFoundError if we try to load a bad pathlib object."""
    with pytest.raises(FileNotFoundError):
        v.data_model.Image(image_data=Path(SAD_IMAGE_PATH))


def test_bad_init_type():
    """Test that we produce a TypeError if we pass a bad type to dm initializer."""

    class SadType:
        pass

    with pytest.raises(TypeError):
        v.data_model.Image(SadType())


### View tests
def test_as_pil():
    """Test that we can grab a view of our DM object as a PIL image."""
    pimg = v.data_model.Image(PIL_IMG_DATA).as_pil()
    # Since we make the dm.Image using PIL as our hub format, the view references the same object
    assert pimg is PIL_IMG_DATA


def test_as_numpy():
    """Test that we can grab a view of our DM object as a Numpy array."""
    dimg = v.data_model.Image(PIL_IMG_DATA)
    expected_shape = (dimg.rows, dimg.columns, dimg.channels)
    npimg = v.data_model.Image(PIL_IMG_DATA).as_numpy()
    assert npimg.shape == expected_shape
    assert npimg.dtype == np.uint8


### Backend validation tests
def test_no_backend_property_access_fails():
    """Test that we throw good errors if we skip _backend setting."""
    sad_img = v.data_model.Image.__new__(v.data_model.Image)
    # Any property relying on a view should raise
    for prop_name in ["rows", "columns", "channels"]:
        with pytest.raises(AttributeError):
            getattr(sad_img, prop_name)
    # So should actually invoking a view method
    for view_name in ["as_numpy", "as_pil"]:
        with pytest.raises(AttributeError):
            getattr(sad_img, view_name)()


def test_from_sad_backend():
    """Test that if garbage is used for our backend, views fail."""
    sad_img = v.data_model.Image.from_backend(100)
    with pytest.raises(AttributeError):
        sad_img.as_numpy()


def test_default_format_pil():
    """Test that we can serialize a PIL image created in-memory with a lossless compression."""
    default_format = "PNG"
    img = PILImage.new("RGB", (1000, 600), color=(255, 255, 255))
    # Image should have no format information
    assert img.format is None
    # We should be able to serialize the image without any issues; we test .proto() since it's used
    # in the underlying conversion to alternate serialized formats, e.g., .json() also.
    dm_image = v.data_model.Image(img)
    proto_img = dm_image.to_proto()
    assert dm_image._backend._export_format == default_format
    # Then, if we reconstruct from the serialized proto, it should be
    # reloaded as a PNG because it was serialized as a PNG.
    assert (
        v.data_model.Image.from_proto(proto_img)._backend._export_format
        == default_format
    )


### Tests for other properties
def test_dm_props():
    """Test that our row, col, and channel properties return the correct values."""
    im = v.data_model.Image(image_data=PIL_IMG_DATA)
    assert im.rows == 100
    assert im.columns == 250
    assert im.channels == 3



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for conversion between python datetime and protobuf Timestamp
"""
# Standard
from typing import List
import datetime

# Third Party
import pytest

# Local
from caikit.core import DataObjectBase, dataobject
from caikit.core.data_model import timestamp
from caikit.interfaces.common.data_model.stream_sources import S3Path
from tests.core.data_model.test_dataobject import auto_reset_global_protobuf_registry


def test_timestamp_to_proto_to_timestamp():
    datetime_ = datetime.datetime.now()
    proto = timestamp.datetime_to_proto(datetime_)
    new_datetime = timestamp.proto_to_datetime(proto)

    assert new_datetime == datetime_


def test_timestamp_to_proto_invalid_type():
    """Make sure that a TypeError is raised if a bad type is encountered"""
    with pytest.raises(TypeError):
        timestamp.datetime_to_proto({"this is not": "a timestamp"})


def test_proto_to_timestamp_invalid_type():
    """Make sure that a TypeError is raised if a bad type is encountered"""
    with pytest.raises(TypeError):
        non_timestamp_proto = S3Path(path="/not/a/timestamp.either")
        timestamp.proto_to_datetime(non_timestamp_proto)


def test_timestamp_roundtrip_serialization():
    @dataobject(package="test")
    class FooTimestamps(DataObjectBase):
        time: datetime.datetime
        times: List[datetime.datetime]

    time_container = FooTimestamps(
        time=datetime.datetime.now(),
        times=[datetime.datetime.now(), datetime.datetime.fromtimestamp(1234567)],
    )

    time_proto = time_container.to_proto()
    time_proto_roundtrip = FooTimestamps.from_proto(time_proto)
    assert time_proto_roundtrip == time_container

    time_json = time_container.to_json()
    time_json_roundtrip = FooTimestamps.from_json(time_json)
    assert time_json_roundtrip == time_container



"""
This sets up global test configs when pytest starts
"""

# Standard
from contextlib import contextmanager
from functools import partial
from typing import Dict, Type, Union
import os
import shlex
import socket
import subprocess
import sys
import tempfile
import threading
import time

# Third Party
from grpc_health.v1 import health_pb2, health_pb2_grpc
import grpc
import pytest
import requests

# First Party
import alog

# Local
from caikit.core import MODEL_MANAGER
from caikit.core.data_model.dataobject import render_dataobject_protos
from caikit.runtime import http_server
from caikit.runtime.grpc_server import RuntimeGRPCServer
from caikit.runtime.model_management.loaded_model import LoadedModel
from caikit.runtime.model_management.model_manager import ModelManager
from caikit.runtime.service_factory import ServicePackage, ServicePackageFactory
from caikit.runtime.service_generation.rpcs import TaskPredictRPC
from caikit.runtime.servicers.global_predict_servicer import GlobalPredictServicer
from caikit.runtime.servicers.global_train_servicer import GlobalTrainServicer
from tests.conftest import random_test_id, temp_config
from tests.fixtures import Fixtures

log = alog.use_channel("TEST-CONFTEST")


@pytest.fixture
def open_port():
    """Get an open port on localhost
    Returns:
        int: Available port
    """
    return _open_port()


@pytest.fixture(scope="session")
def session_scoped_open_port():
    """Get an open port on localhost
    Returns:
        int: Available port
    """
    return _open_port()


@pytest.fixture(scope="session")
def http_session_scoped_open_port():
    """Get an open port on localhost
    Returns:
        int: Available port
    """
    return _open_port()


def _open_port():
    # TODO: This has obvious problems where the port returned for use by a test is not immediately
    # put into use, so parallel tests could attempt to use the same port.
    start = 8888
    end = start + 1000
    host = "localhost"
    for port in range(start, end):
        with socket.socket() as soc:
            # soc.connect_ex returns 0 if connection is successful,
            # indicating the port is in use
            if soc.connect_ex((host, port)) != 0:
                # So a non-zero code should mean the port is not currently in use
                return port


@pytest.fixture(scope="session")
def sample_inference_service(render_protos) -> ServicePackage:
    """Service package pointing to `sample_lib` for testing"""
    inference_service = ServicePackageFactory().get_service_package(
        ServicePackageFactory.ServiceType.INFERENCE,
    )
    if render_protos:
        output_dir = os.path.join("tests", "protos")
        render_dataobject_protos(output_dir)
        inference_service.service.write_proto_file(output_dir)
    return inference_service


@pytest.fixture(scope="session")
def sample_predict_servicer(sample_inference_service) -> GlobalPredictServicer:
    servicer = GlobalPredictServicer(inference_service=sample_inference_service)
    yield servicer
    # Make sure to not leave the rpc_meter hanging
    # (It does try to clean itself up on destruction, but just to be sure)
    rpc_meter = getattr(servicer, "rpc_meter", None)
    if rpc_meter:
        rpc_meter.end_writer_thread()


@pytest.fixture(scope="session")
def sample_train_service(render_protos) -> ServicePackage:
    """Service package pointing to `sample_lib` for testing"""
    training_service = ServicePackageFactory().get_service_package(
        ServicePackageFactory.ServiceType.TRAINING,
    )
    if render_protos:
        output_dir = os.path.join("tests", "protos")
        render_dataobject_protos(output_dir)
        training_service.service.write_proto_file(output_dir)
    return training_service


@pytest.fixture(scope="session")
def sample_train_servicer(sample_train_service) -> GlobalTrainServicer:
    servicer = GlobalTrainServicer(training_service=sample_train_service)
    yield servicer


@contextmanager
def runtime_grpc_test_server(open_port, *args, **kwargs):
    """Helper to wrap creation of RuntimeGRPCServer in temporary configurations"""
    with tempfile.TemporaryDirectory() as workdir:
        temp_log_dir = os.path.join(workdir, "metering_logs")
        temp_save_dir = os.path.join(workdir, "training_output")
        os.makedirs(temp_log_dir)
        os.makedirs(temp_save_dir)
        with temp_config(
            {
                "runtime": {
                    "metering": {"log_dir": temp_log_dir},
                    "training": {"output_dir": temp_save_dir},
                    "grpc": {"port": open_port},
                }
            },
            "merge",
        ):
            with RuntimeGRPCServer(*args, **kwargs) as server:
                # Give tests access to the workdir
                server.workdir = workdir
                yield server


@pytest.fixture(scope="session")
def runtime_grpc_server(
    session_scoped_open_port,
) -> RuntimeGRPCServer:
    with runtime_grpc_test_server(
        session_scoped_open_port,
    ) as server:
        _check_server_readiness(server)
        yield server


@contextmanager
def runtime_http_test_server(open_port, *args, **kwargs):
    """Helper to wrap creation of RuntimeHTTPServer in temporary configurations"""
    with tempfile.TemporaryDirectory() as workdir:
        temp_log_dir = os.path.join(workdir, "metering_logs")
        temp_save_dir = os.path.join(workdir, "training_output")
        os.makedirs(temp_log_dir)
        os.makedirs(temp_save_dir)
        with temp_config(
            {
                "runtime": {
                    "metering": {"log_dir": temp_log_dir},
                    "training": {"output_dir": temp_save_dir},
                    "http": {"port": open_port},
                }
            },
            "merge",
        ):
            config_overrides = {}
            if "tls_config_override" in kwargs:
                config_overrides = kwargs["tls_config_override"]
                kwargs["tls_config_override"] = config_overrides["runtime"]["tls"]
            with http_server.RuntimeHTTPServer(*args, **kwargs) as server:
                _check_http_server_readiness(server, config_overrides)
                # Give tests access to the workdir
                server.workdir = workdir
                yield server


# For tests that need a working http server
# For other http tests, we can use FastAPI TestClient
@pytest.fixture(scope="session")
def runtime_http_server(
    http_session_scoped_open_port,
) -> http_server.RuntimeHTTPServer:
    """yields an actual running http server"""
    with runtime_http_test_server(
        http_session_scoped_open_port,
    ) as server:
        yield server


@pytest.fixture(scope="session")
def inference_stub(sample_inference_service, runtime_grpc_server) -> Type:
    inference_stub = sample_inference_service.stub_class(
        runtime_grpc_server.make_local_channel()
    )
    return inference_stub


@pytest.fixture(scope="session")
def train_stub(sample_train_service, runtime_grpc_server) -> Type:
    train_stub = sample_train_service.stub_class(
        runtime_grpc_server.make_local_channel()
    )
    return train_stub


@pytest.fixture(scope="session")
def training_management_stub(runtime_grpc_server) -> Type:
    training_management_service: ServicePackage = (
        ServicePackageFactory().get_service_package(
            ServicePackageFactory.ServiceType.TRAINING_MANAGEMENT,
        )
    )

    training_management_stub = training_management_service.stub_class(
        runtime_grpc_server.make_local_channel()
    )
    return training_management_stub


@pytest.fixture
def sample_task_model_id(good_model_path) -> str:
    """Loaded model ID using model manager load model implementation"""
    model_id = random_test_id()
    model_manager = ModelManager.get_instance()
    # model load test already tests with archive - just using a model path here
    model_manager.load_model(
        model_id,
        local_model_path=good_model_path,
        model_type=Fixtures.get_good_model_type(),  # eventually we'd like to be determining the type from the model itself...
    )
    yield model_id

    # teardown
    model_manager.unload_model(model_id)


@pytest.fixture
def file_task_model_id(box_model_path) -> str:
    """Loaded model ID using model manager load model implementation"""
    model_id = random_test_id()
    model_manager = ModelManager.get_instance()
    # model load test already tests with archive - just using a model path here
    model_manager.load_model(
        model_id,
        local_model_path=box_model_path,
        model_type=Fixtures.get_good_model_type(),  # eventually we'd like to be determining the type from the model itself...
    )
    yield model_id

    # teardown
    model_manager.unload_model(model_id)


@pytest.fixture
def sample_task_unary_rpc(sample_inference_service: ServicePackage) -> TaskPredictRPC:
    return sample_inference_service.caikit_rpcs["SampleTaskPredict"]


@pytest.fixture
def sample_task_streaming_rpc(
    sample_inference_service: ServicePackage,
) -> TaskPredictRPC:
    return sample_inference_service.caikit_rpcs["ServerStreamingSampleTaskPredict"]


@pytest.fixture
def streaming_task_model_id(streaming_model_path) -> str:
    """Loaded model ID using model manager load model implementation"""
    model_id = random_test_id()
    model_manager = ModelManager.get_instance()
    model_manager.load_model(
        model_id,
        local_model_path=streaming_model_path,
        model_type=Fixtures.get_good_model_type(),
    )
    yield model_id

    # teardown
    model_manager.unload_model(model_id)


@pytest.fixture
def other_task_model_id(other_good_model_path) -> str:
    """Loaded model ID using model manager load model implementation"""
    model_id = random_test_id()
    model_manager = ModelManager.get_instance()
    # model load test already tests with archive - just using a model path here
    model_manager.load_model(
        model_id,
        local_model_path=other_good_model_path,
        model_type=Fixtures.get_good_model_type(),  # eventually we'd like to be determining the type from the model itself...
    )
    yield model_id

    # teardown
    model_manager.unload_model(model_id)


def register_trained_model(
    servicer: Union[RuntimeGRPCServer, GlobalPredictServicer],
    model_id: str,
    training_id: str,
):
    """Helper to auto-load a model that has completed training. This replaces
    the old auto-load feature which was only needed for unit tests
    """
    model_future_factory = partial(MODEL_MANAGER.get_model_future, training_id)
    loaded_model = (
        LoadedModel.Builder()
        .id(model_id)
        .type("trained")
        .path("")
        .model_future_factory(model_future_factory)
        .build()
    )
    if isinstance(servicer, RuntimeGRPCServer):
        servicer = servicer._global_predict_servicer
    servicer._model_manager.loaded_models[model_id] = loaded_model


class ModuleSubproc:
    def __init__(
        self, module_to_run: str, *args, kill_timeout: float = 10.0, **env_vars
    ):
        """Run the given python as a subprocess and kill it after the given timeout"""
        # Set up the command
        cmd = f"{sys.executable} -m {module_to_run}"
        for arg in args:
            cmd += f" {arg}"
        self._cmd = shlex.split(cmd)

        # Set up the environment
        self._env = {**os.environ, **env_vars}
        self._env["PYTHONPATH"] = ":".join(sys.path)

        # Start the process
        self.proc = None
        self._killed = False

        # Start the timer that will kill the process
        self._kill_timer = threading.Timer(kill_timeout, self._kill_proc)

    @property
    def killed(self):
        return self._killed

    def _kill_proc(self):
        self._killed = True
        if self.proc is not None:
            self.proc.kill()

    def __enter__(self):
        self.proc = subprocess.Popen(self._cmd, env=self._env)
        self._kill_timer.start()
        return self.proc

    def __exit__(self, *_):
        self.proc.wait()
        self._kill_timer.cancel()
        self._kill_timer.join()


# IMPLEMENTATION DETAILS ############################################################


def _check_server_readiness(server):
    """Check server readiness"""

    channel = grpc.insecure_channel(f"localhost:{server.port}")

    done = False
    while not done:
        try:
            stub = health_pb2_grpc.HealthStub(channel)
            health_check_request = health_pb2.HealthCheckRequest()
            stub.Check(health_check_request)
            done = True
        except grpc.RpcError:
            log.debug(
                "[RpcError]; will try to reconnect to test server in 0.01 second."
            )
            time.sleep(0.01)


def _check_http_server_readiness(server, config_overrides: Dict[str, Dict]):
    mode = "http"
    verify = None
    cert = None
    # tls
    if config_overrides:
        mode = "https"
        verify = config_overrides["use_in_test"]["ca_cert"]
        # mtls
        if "client_cert" and "client_key" in config_overrides["use_in_test"]:
            cert = (
                config_overrides["use_in_test"]["client_cert"],
                config_overrides["use_in_test"]["client_key"],
            )
    done = False
    while not done:
        try:
            response = requests.get(
                f"{mode}://localhost:{server.port}{http_server.HEALTH_ENDPOINT}",
                verify=verify,
                cert=cert,
            )
            assert response.status_code == 200
            assert response.text == "OK"
            done = True
        except AssertionError:
            log.debug(
                "[HTTP server not ready]; will try to reconnect to test server in 0.01 second."
            )
            time.sleep(0.001)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
import threading
import unittest

# Local
from caikit.runtime.work_management.rpc_aborter import RpcAborter
from tests.fixtures import Fixtures


class TestRpcAborter(unittest.TestCase):
    """This test suite tests the call aborter utility"""

    CHANNEL = None
    GRPC_THREAD = None

    def test_call_aborter_sets_event(self):
        ctx = Fixtures.build_context("call_aborter_event_party")
        # Create a new Call aborter
        aborter = RpcAborter(ctx)
        # Create a new threading event and add it to the call aborter
        event = threading.Event()
        aborter.add_event(event)
        # Cancel the call & wait for the threading event to be set by __rpc_terminated
        ctx.cancel()
        event.wait()
        self.assertTrue(aborter.must_abort())

    def test_call_aborter_sets_event_added_after_termination(self):
        ctx = Fixtures.build_context("call_aborter_event_party")
        # Create a new call aborter
        aborter = RpcAborter(ctx)
        # Cancel the call before creating the threading event and adding to the aborter
        ctx.cancel()
        event = threading.Event()
        aborter.add_event(event)
        event.wait()
        self.assertTrue(aborter.must_abort())


if __name__ == "__main__":
    unittest.main()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from typing import Optional
import ctypes
import sys
import threading
import traceback

# First Party
import alog

# Local
from .destroyable import Destroyable

log = alog.use_channel("DESTROY-THRD")


class ThreadDestroyedException(RuntimeError):
    """Exception raised inside a DestroyableThread when it is destroyed by the thread managing
    its lifecycle."""

    def __init__(self):
        super().__init__(
            "Work thread intentionally destroyed by its lifecycle manager. "
            "This exception was not raised by the code running in this thread."
        )


# pylint: disable=too-many-instance-attributes
class DestroyableThread(threading.Thread, Destroyable):
    """A class for Destroyable Threads. When work is delegated to a thread but may need to be
    canceled while in progress, we use this class which allows us to raise an exception inside
    the work thread.

    Exceptions raised this way are asynchronous and they will not interrupt the python instruction
    that the thread is currently executing. E.g. a time.sleep() will finish sleeping before the
    exception is raised.

    This class may be initialized with a threading event, which it will set when the thread
    finishes executing, whether nominally or by raising an exception.
    """

    # The exception we'll throw to kill the thread
    __exception = ThreadDestroyedException

    def __init__(
        self,
        runnable_func,
        *runnable_args,
        work_done_event: Optional[threading.Event] = None,
        **runnable_kwargs
    ):
        threading.Thread.__init__(self)
        self.work_done_event = work_done_event or threading.Event()

        # These describe the work to be done
        self.runnable_func = runnable_func
        self.runnable_args = runnable_args
        self.runnable_kwargs = runnable_kwargs

        # These describe what happened with the work
        self.__runnable_result = None
        self.__runnable_exception = None
        self.__threw = False
        self.__started = False
        self.__ran = False

        # In case `destroy` is called before Python has actually started the thread, we need to
        # know to not do the work
        self.__destroyed = False

    @property
    def destroyed(self) -> bool:
        return self.__destroyed

    @property
    def canceled(self) -> bool:
        return self.destroyed and (
            (self.__started and (self.threw or not self.ran)) or not self.__started
        )

    @property
    def ran(self) -> bool:
        return self.__ran

    @property
    def threw(self) -> bool:
        return self.__threw

    # Run wraps the supplied function with logic to set the event when it finishes, and save any
    # result or raised error

    def run(self) -> None:
        """
            Overrides Thread.run()
            *Do not call*

        Returns:
            None
        """
        # Raise immediately if the thread was destroyed before
        if self.__destroyed:
            log.info(
                "<COR14653273I>",
                "Not starting work for %s, thread already cancelled",
                self.runnable_func,
            )
            self.__raise()

        self.__started = True
        try:
            self.__runnable_result = self.runnable_func(
                *self.runnable_args, **self.runnable_kwargs
            )
            self.__threw = False
        except:  # pylint: disable=bare-except
            # PEP8 complains, but in this case we really do want to re-throw _any_ exception that
            # occurred. In the interest of transparently wrapping any work in these threads, we
            # want to keep exception signatures identical. E.g. if I expect this thread to throw a
            # CaikitRuntimeException, I want to be able to catch a CaikitRuntimeException.
            # Rethrowing from `sys.exc_info()[1]` should retain all stack trace info later.
            e = sys.exc_info()[1]
            self.__runnable_exception = e
            self.__threw = True

            # Add a little bit of visibility to know why work failed
            if self.__destroyed:
                log.info(
                    {
                        "log_code": "<COR15827563I>",
                        "message": "Work for {} was aborted and threw".format(
                            self.runnable_func
                        ),
                        "stack_trace": traceback.format_exc(),
                    }
                )
            else:
                log.warning(
                    {
                        "log_code": "<COR16788843W>",
                        "message": "Work for {} threw exception: {}".format(
                            self.runnable_func, e
                        ),
                        "stack_trace": traceback.format_exc(),
                    }
                )

        finally:
            # Before setting the synchronization event, flag that the work was done
            self.__ran = True
            self.work_done_event.set()

    def get_or_throw(self):
        """
            After the thread has completed it's work, call this to get the output.
        Returns:
            The resulting value of runnable_func(*runnable_args, **runnable_kwargs)
        Raises:
            Any exception raised by runnable_func(*runnable_args, **runnable_kwargs)
        """
        if self.destroyed:
            log.error(
                "<COR14653274E>",
                "get_or_throw called on destroyed thread for %s, no value to return",
                self.runnable_func,
            )

        if not self.ran:
            log.error(
                "<COR14653275E>",
                "get_or_throw called on thread for %s, but it has not finished running",
                self.runnable_func,
            )

        if self.threw:
            raise self.__runnable_exception
        return self.__runnable_result

    def destroy(self) -> None:
        """
            Cancel any in-progress work and kill the thread if it is alive.
            Otherwise, prevent the thread from running at all.
        Returns:
            None
        """

        # Set the destroyed flag in case the thread has not started yet.
        # If it has, we should be able to kill it with the async exception below.
        self.__destroyed = True

        # The thread has already finished or is not yet alive, so we cannot kill it
        thread_id = self.__get_id()
        if thread_id is None or not self.is_alive():
            log.debug(
                "<COR14653276D>",
                "Destroying thread that is not currently alive: %s",
                self.runnable_func,
            )
            return

        # This is the code that raises an async exception in the target thread
        # (We can't just use raise, because the parent thread is in this control flow)
        async_exception_result = ctypes.pythonapi.PyThreadState_SetAsyncExc(
            ctypes.c_long(thread_id), ctypes.py_object(self.__exception)
        )
        if async_exception_result > 1:
            log.error(
                "<COR14653277E>",
                "Could not raise async exception on destroyable thread for %s. Result code: %s",
                self.runnable_func,
                async_exception_result,
            )

    @property
    def error(self) -> Optional[Exception]:
        if isinstance(self.__runnable_exception, Exception):
            return self.__runnable_exception

    def __get_id(self):
        # Returns the thread if if the thread is running
        for thread in threading.enumerate():
            if thread is self:
                return thread.ident

        # Otherwise, the thread has completed or has not started
        return None

    def __raise(self):
        # __exception is just a type, we need to be sure to initialize a value of it
        raise self.__exception()
